import functools
import json
from pathlib import Path
from typing import Dict

import numpy as np

from pysisyphus.Geometry import Geometry


@functools.singledispatch
def geom_from_qcschema(qcschema: Dict, **geom_kwargs):
    mol = qcschema["molecule"]
    coords = np.array((mol["geometry"]))
    atoms = mol["symbols"]
    return Geometry(atoms, coords, **geom_kwargs)


# These definitions may seem a bit wild, but I was not able to make
# singledispatch work with 'file_or_str'.


@geom_from_qcschema.register
def _(text: str, **geom_kwargs):
    if (p := Path(text).exists()):
        with open(p, "r") as handle:
            text = handle.read()
    qcschema = json.loads(text)
    return geom_from_qcschema(qcschema, **geom_kwargs)


@geom_from_qcschema.register
def _(path: Path, **geom_kwargs):
    return geom_from_qcschema(str(path), **geom_kwargs)



"""CLI-level bool compatibility regression tests."""

from __future__ import annotations

import click
from click.testing import CliRunner

from pdb2reaction.cli import cli as root_cli


def _is_unavailable_command(cmd: click.Command | None) -> bool:
    if cmd is None:
        return True
    help_text = (cmd.help or "").strip()
    return help_text.startswith("[Unavailable]")


def test_all_bool_options_accept_toggle_and_value_styles() -> None:
    runner = CliRunner()
    ctx = click.Context(root_cli)

    tested = 0
    for command_name in root_cli.list_commands(ctx):
        command = root_cli.get_command(ctx, command_name)
        if _is_unavailable_command(command):
            continue

        value_opts, toggle_opts, _aliases, single_opts = root_cli._resolve_bool_options(
            ctx, command_name
        )
        option_names = sorted(
            {
                *value_opts,
                *toggle_opts,
                *single_opts,
            }
        )
        for opt in option_names:
            if not opt.startswith("--"):
                continue

            no_opt = f"--no-{opt[2:]}"

            result_flag = runner.invoke(root_cli, [command_name, opt, "--help"])
            assert result_flag.exit_code == 0, (
                f"{command_name} should accept {opt}. Output:\n{result_flag.output}"
            )

            result_no_flag = runner.invoke(root_cli, [command_name, no_opt, "--help"])
            assert result_no_flag.exit_code == 0, (
                f"{command_name} should accept {no_opt}. Output:\n{result_no_flag.output}"
            )

            result_value_false = runner.invoke(
                root_cli, [command_name, opt, "False", "--help"]
            )
            assert result_value_false.exit_code == 0, (
                f"{command_name} should accept '{opt} False'. Output:\n"
                f"{result_value_false.output}"
            )

            tested += 1

    assert tested > 0



import logging

from pysisyphus.tsoptimizers.RSPRFOptimizer import RSPRFOptimizer
from pysisyphus.tsoptimizers.TRIM import TRIM
from pysisyphus.tsoptimizers.RSIRFOptimizer import RSIRFOptimizer

__all__ = [
    "RSPRFOptimizer",
    "TRIM",
    "RSIRFOptimizer",
    "TSHessianOptimizer",
]


logger = logging.getLogger("tsoptimizer")
logger.setLevel(logging.DEBUG)
# delay = True prevents creation of empty logfiles
handler = logging.FileHandler("tsoptimizer.log", mode="w", delay=True)
# fmt_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
fmt_str = "%(message)s"
formatter = logging.Formatter(fmt_str)
handler.setFormatter(formatter)
logger.addHandler(handler)



from pysisyphus.optimizers import (
    BFGS,
    CubicNewton,
    ConjugateGradient,
    FIRE,
    LayerOpt,
    LBFGS,
    MicroOptimizer,
    NCOptimizer,
    PreconLBFGS,
    PreconSteepestDescent,
    QuickMin,
    RFOptimizer,
    SteepestDescent,
    StabilizedQNMethod,
    StringOptimizer,
)
from pysisyphus.tsoptimizers import (
    RSPRFOptimizer,
    TRIM,
    RSIRFOptimizer,
)


OPT_DICT = {
    "bfgs": BFGS.BFGS,
    "cg": ConjugateGradient.ConjugateGradient,
    "cnewton": CubicNewton,
    "fire": FIRE.FIRE,
    "layers": LayerOpt.LayerOpt,
    "lbfgs": LBFGS.LBFGS,
    "micro": MicroOptimizer,
    "nc": NCOptimizer.NCOptimizer,
    "plbfgs": PreconLBFGS.PreconLBFGS,
    "psd": PreconSteepestDescent.PreconSteepestDescent,
    "qm": QuickMin.QuickMin,
    "rfo": RFOptimizer.RFOptimizer,
    "sd": SteepestDescent.SteepestDescent,
    "sqnm": StabilizedQNMethod.StabilizedQNMethod,
    "string": StringOptimizer.StringOptimizer,
}

TSOPT_DICT = {
    "rsprfo": RSPRFOptimizer,
    "prfo": RSPRFOptimizer,  # Shortcut
    "trim": TRIM,
    "rsirfo": RSIRFOptimizer,
    "irfo": RSIRFOptimizer,  # Shortcut
}
OPT_DICT.update(TSOPT_DICT)


def get_opt_cls(opt_key):
    return OPT_DICT[opt_key]


def key_is_tsopt(opt_key):
    return opt_key in TSOPT_DICT



from pysisyphus.calculators.AnaPotBase import AnaPotBase

# See J. Chem. Phys. 122 174106 (2005)
# https://doi.org/10.1063/1.1885467
# Eq. (11)

class FourWellAnaPot(AnaPotBase):

    def __init__(self): 
        V_str = "x**4 + y**4 - 2*x**2 - 4*y**2 + x*y + 0.3*x + 0.1*y"
        xlim = (-1.75, 1.75)
        ylim = (-1.75, 1.75)
        minima = (
            (1.12410175, -1.48527428, 0.0),
            (-0.82190767, -1.36672971, 0.0),
            (-1.17405609,  1.47708706, 0.0),
        )
        super().__init__(V_str=V_str, xlim=xlim, ylim=ylim, minima=minima)

    def __str__(self):
        return "FourWellAnaPot calculator"


if __name__ == "__main__":
    fw = FourWellAnaPot()
    fw.plot()
    import matplotlib.pyplot as plt
    plt.show()



from math import sin

import numpy as np

from pysisyphus.intcoords.Primitive import Primitive
from pysisyphus.intcoords import Bend
from pysisyphus.intcoords.derivatives import d2q_d2
from pysisyphus.linalg import cross3, norm3


class Torsion(Primitive):
    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        m, o, p, n = indices
        rho_mo = Torsion.rho(atoms, coords3d, (m, o))
        rho_op = Torsion.rho(atoms, coords3d, (o, p))
        rho_pn = Torsion.rho(atoms, coords3d, (p, n))
        rad_mop = Bend._calculate(coords3d, (m, o, p))
        rad_opn = Bend._calculate(coords3d, (o, p, n))
        return (
            (rho_mo * rho_op * rho_pn) ** (1 / 3)
            * (f_damping + (1 - f_damping) * sin(rad_mop))
            * (f_damping + (1 - f_damping) * sin(rad_opn))
        )

    @staticmethod
    def _calculate(coords3d, indices, gradient=False):
        m, o, p, n = indices
        u_dash = coords3d[m] - coords3d[o]
        v_dash = coords3d[n] - coords3d[p]
        w_dash = coords3d[p] - coords3d[o]
        u_norm = norm3(u_dash)
        v_norm = norm3(v_dash)
        w_norm = norm3(w_dash)
        u = u_dash / u_norm
        v = v_dash / v_norm
        w = w_dash / w_norm
        phi_u = np.arccos(u.dot(w))
        phi_v = np.arccos(-w.dot(v))
        uxw = cross3(u, w)
        vxw = cross3(v, w)
        cos_dihed = uxw.dot(vxw) / (np.sin(phi_u) * np.sin(phi_v))
        # Restrict cos_dihed to the allowed interval for arccos [-1, 1]
        cos_dihed = min(1, max(cos_dihed, -1))

        dihedral_rad = np.arccos(cos_dihed)

        # Arccos only returns values between 0 and π, but dihedrals can
        # also be negative. This is corrected now.
        #
        # (v ⨯ w) · u will be < 0 when both vectors point in different directions.
        #
        #  M  --->   N
        #  ^        ^
        #   \      /
        #    u    v    positive dihedral, M rotates into N clockwise
        #     \  /     (v ⨯ w) · u > 0, keep positive sign
        #      OwP
        #              w points downward, into the screen plane.
        #              The vector resulting from the cross-product is easily
        #              visualized with your right hand.
        #
        #  M
        #   \
        #  | u
        #  |  \
        #  |   OwP     negative dihedral, M rotates into N counter-clockwise
        #  v  /        (v ⨯ w) · u < 0, invert dihedral sign
        #    v
        #   /
        #  N
        #
        if (dihedral_rad != np.pi) and (vxw.dot(u) < 0):
            dihedral_rad *= -1

        if gradient:
            row = np.zeros_like(coords3d)
            #                  |  m  |  n  |  o  |  p  |
            # ------------------------------------------
            # sign_factor(amo) |  1  |  0  | -1  |  0  | 1st term
            # sign_factor(apn) |  0  | -1  |  0  |  1  | 2nd term
            # sign_factor(aop) |  0  |  0  |  1  | -1  | 3rd term
            # sign_factor(apo) |  0  |  0  | -1  |  1  | 4th term
            sin2_u = np.sin(phi_u) ** 2
            sin2_v = np.sin(phi_v) ** 2
            first_term = uxw / (u_norm * sin2_u)
            second_term = vxw / (v_norm * sin2_v)
            third_term = uxw * np.cos(phi_u) / (w_norm * sin2_u)
            fourth_term = -vxw * np.cos(phi_v) / (w_norm * sin2_v)
            row[m, :] = first_term
            row[n, :] = -second_term
            row[o, :] = -first_term + third_term - fourth_term
            row[p, :] = second_term - third_term + fourth_term
            row = row.flatten()
            return dihedral_rad, row
        return dihedral_rad

    @staticmethod
    def _jacobian(coords3d, indices):
        sign = np.sign(Torsion._calculate(coords3d, indices))
        return sign * d2q_d2(*coords3d[indices].flatten())



# [1] https://aip.scitation.org/doi/pdf/10.1063/1.1523908
#     Neugebauer, Reiher 2002
# [2] https://reiher.ethz.ch/software/akira.html


from collections import namedtuple
import sys

import numpy as np

from pysisyphus.helpers_pure import eigval_to_wavenumber
from pysisyphus.Geometry import get_trans_rot_projector
from pysisyphus.modefollow.NormalMode import NormalMode
from pysisyphus.TablePrinter import TablePrinter


DavidsonResult = namedtuple(
    "DavidsonResult",
    "cur_cycle converged final_modes qs nus mode_inds res_rms",
)


def forces_fin_diff(forces_getter, coords, b, step_size):
    plus = forces_getter(coords + b)
    minus = forces_getter(coords - b)
    fd = (minus - plus) / (2 * step_size)
    return fd


def block_davidson(
    cart_coords,
    masses,
    forces_getter,
    guess_modes,
    lowest=None,
    trial_step_size=0.01,
    hessian_precon=None,
    max_cycles=25,
    res_rms_thresh=1e-4,
    start_precon=5,
    remove_trans_rot=True,
    print_level=1,
):
    num = len(guess_modes)
    B_full = np.zeros((len(guess_modes[0]), num * max_cycles))
    S_full = np.zeros_like(B_full)
    I = np.eye(cart_coords.size)
    masses_rep = np.repeat(masses, 3)
    msqrt = np.sqrt(masses_rep)

    # Projector to remove translation and rotation
    P = get_trans_rot_projector(cart_coords, masses, full=True)
    guess_modes = [
        NormalMode(P.dot(mode.l_mw) / msqrt, masses_rep) for mode in guess_modes
    ]

    col_fmts = "int int int float_short float str".split()
    header = ("#", "subspace size", "mode", "ṽ / cm⁻¹", "rms(r)", "Conv")
    fmts_update = {"float_short": "{: >11.2f}"}
    table = TablePrinter(header, col_fmts, width=11, fmts_update=fmts_update)
    if print_level == 1:
        table.print_header()

    b_prev = np.array([mode.l_mw for mode in guess_modes]).T
    for i in range(max_cycles):
        # Add new basis vectors to B matrix
        b = np.array([mode.l_mw for mode in guess_modes]).T
        from_ = i * num
        to_ = (i + 1) * num
        B_full[:, from_:to_] = b

        # Estimate action of Hessian by finite differences.
        for j in range(num):
            mode = guess_modes[j]
            # Get a step size in mass-weighted coordinates that results
            # in the desired 'trial_step_size' in not-mass-weighted coordinates.
            mw_step_size = mode.mw_norm_for_norm(trial_step_size)
            # Actual step in non-mass-weighted coordinates
            step = trial_step_size * mode.l
            S_full[:, from_ + j] = (
                # Convert to mass-weighted coordinates
                forces_fin_diff(forces_getter, cart_coords, step, mw_step_size)
                / msqrt
            )

        # Views on columns that are actually set
        B = B_full[:, :to_]
        S = S_full[:, :to_]

        # Calculate and symmetrize approximate hessian
        Hm = B.T.dot(S)
        Hm = (Hm + Hm.T) / 2
        # Diagonalize small Hessian
        w, v = np.linalg.eigh(Hm)

        # Approximations to exact eigenvectors in current cycle
        approx_modes = (v * B[:, :, None]).sum(axis=1).T
        # Calculate overlaps between previous root and the new approximate
        # normal modes for root following.
        if lowest is None:
            # 2D overlap array. approx_modes in row, b_prev in columns.
            overlaps = np.einsum("ij,jk->ik", approx_modes, b_prev)
            mode_inds = np.abs(overlaps).argmax(axis=0)
        else:
            mode_inds = np.arange(lowest)
        b_prev = approx_modes[mode_inds].T

        # Eq. (7) in [1]
        residues = (v * (S[:, :, None] - w * B[:, :, None])).sum(axis=1)

        # Determine preconditioner matrix
        #
        # Use supplied matrix
        if hessian_precon is not None:
            precon_mat = hessian_precon
        # Reconstruct Hessian, but only start after some cycles
        elif i >= start_precon:
            precon_mat = B.dot(Hm).dot(B.T)
        # No preconditioning if no matrix was supplied or we are in an early cycle.
        else:
            precon_mat = None

        # Construct new basis vector from residuum of selected mode
        b = np.zeros_like(b_prev)
        for j, mode_ind in enumerate(mode_inds):
            r = residues[:, mode_ind]
            if precon_mat is not None:
                # Construct actual preconditioner X
                X = np.linalg.pinv(precon_mat - w[mode_ind] * I, rcond=1e-8)
                b[:, j] = X.dot(r)
            else:
                b[:, j] = r

        # Project out translation and rotation from new mode guess
        if remove_trans_rot:
            b = P.dot(b)
        # Orthogonalize new vectors against preset vectors
        b = np.linalg.qr(np.concatenate((B, b), axis=1))[0][:, -num:]

        # New NormalMode from non-mass-weighted displacements
        guess_modes = [NormalMode(b_ / msqrt, masses_rep) for b_ in b.T]

        # Calculate wavenumbers
        nus = eigval_to_wavenumber(w)

        # Check convergence criteria
        max_res = np.abs(residues).max(axis=0)
        res_rms = np.sqrt(np.mean(residues ** 2, axis=0))

        converged = res_rms < res_rms_thresh
        # Print progress if requested
        if print_level == 2:
            print(f"Cycle {i:02d}")
            print("\t #  |    ṽ / cm⁻¹|   rms(r)   | max(|r|) ")
            print("\t------------------------------------------")
            for j, (nu, rms, mr) in enumerate(zip(nus, res_rms, max_res)):
                sel_str = "*" if (j in mode_inds) else " "
                conv_str = "✓" if converged[j] else ""
                print(
                    f"\t{j:02d}{sel_str} | {nu:> 10.2f} | {rms:.8f} | {mr:.8f} {conv_str}"
                )
            print()
        elif print_level == 1:
            for j in mode_inds:
                conv_str = "✓" if converged[j] else "✗"
                table.print_row((i, B.shape[1], j, nus[j], res_rms[j], conv_str))

        # Convergence is signalled using only the roots we are actually interested in
        modes_converged = all(converged[mode_inds])
        if modes_converged:
            if print_level > 0:
                print(f"\tDavidson procedure converged in {i+1} cycles!")
                if lowest is not None:
                    nus_str = np.array2string(nus[mode_inds], precision=2)
                    print(f"\tLowest {lowest} wavenumbers: {nus_str} cm⁻¹")
                    neg_nus = sum(nus[mode_inds] < 0)
                    type_ = "minimum" if (neg_nus == 0) else f"saddle point of index {neg_nus}"
                    print(f"\tThis geometry seems to be a {type_} on the PES.")
            break
        sys.stdout.flush()

    result = DavidsonResult(
        cur_cycle=i,
        converged=modes_converged,
        final_modes=guess_modes,
        qs=approx_modes,
        nus=nus,
        mode_inds=mode_inds,
        res_rms=res_rms,
    )

    return result


def geom_davidson(geom, *args, **kwargs):
    def forces_getter(cart_coords):
        return geom.get_energy_and_cart_forces_at(cart_coords)["forces"]

    return block_davidson(geom.cart_coords, geom.masses, forces_getter, *args, **kwargs)



# pdb2reaction/backends/solvent.py

"""
xTB implicit-solvent delta-correction wrapper.

Wraps any ``MLIPCalculator`` and adds an xTB-based implicit-solvent
correction::

    ΔE = E_xTB(solvent) − E_xTB(vacuum)
    ΔF = F_xTB(solvent) − F_xTB(vacuum)
    ΔH = H_xTB(solvent) − H_xTB(vacuum)

The correction is added to the MLIP energy/forces/Hessian.

Usage::

    from pdb2reaction.backends.solvent import SolventCorrectedCalculator

    base = create_calculator(backend="uma", charge=0, spin=1)
    calc = SolventCorrectedCalculator(base, solvent="water", solvent_model="alpb")
"""

from __future__ import annotations

from typing import Any, Dict, Optional, Sequence

import numpy as np

from pysisyphus.constants import BOHR2ANG, AU2EV

from .base import MLIPCalculator, BackendError

# Unit conversion (atomic units ↔ eV/Å)
_EV2AU = 1.0 / AU2EV
_ANG2BOHR = 1.0 / BOHR2ANG
_F_EVAA_2_AU = _EV2AU / _ANG2BOHR
_H_EVAA_2_AU = _EV2AU / _ANG2BOHR / _ANG2BOHR

# xTB correction helpers (lazy-imported from bundled module)
_xtb_mod = None


def _get_xtb_mod():
    global _xtb_mod
    if _xtb_mod is None:
        from . import xtb_alpb_correction as _mod
        _xtb_mod = _mod
    return _xtb_mod


def solvent_correction_enabled(solvent: str) -> bool:
    """Return True if solvent is not 'none'/empty."""
    mod = _get_xtb_mod()
    return mod.solvent_correction_enabled(solvent)


class SolventCorrectedCalculator(MLIPCalculator):
    """Decorator that adds xTB implicit-solvent correction to any MLIP calculator.

    If ``solvent`` is ``'none'`` (default), all calls are passed through
    to the base calculator unchanged.
    """

    def __init__(
        self,
        base: MLIPCalculator,
        *,
        solvent: str = "none",
        solvent_model: str = "alpb",
        xtb_cmd: str = "xtb",
        xtb_acc: float = 0.2,
    ):
        # Inherit charge/spin/device from base
        super().__init__(
            charge=base.charge,
            spin=base.mult,
            device=getattr(base, "device_str", "auto"),
            freeze_atoms=base.freeze_atoms,
            hessian_calc_mode=base.hessian_calc_mode,
            return_partial_hessian=base.return_partial_hessian,
            hessian_double=base.hessian_double,
            print_timing=base.print_timing,
        )

        self.base = base
        self.solvent = str(solvent)
        self.solvent_model = str(solvent_model)
        self.xtb_cmd = str(xtb_cmd)
        self.xtb_acc = float(xtb_acc)

        self._enabled = solvent_correction_enabled(self.solvent)

        if self._enabled:
            mod = _get_xtb_mod()
            # Validate solvent model
            mod.normalize_solvent_model(self.solvent_model)

    def _solvent_delta(
        self,
        symbols: Sequence[str],
        coords_ang: np.ndarray,
        need_forces: bool = False,
        need_hessian: bool = False,
    ):
        """Compute xTB solvent delta (eV/Å units)."""
        mod = _get_xtb_mod()
        de_ev, df_ev_ang, dh_ev_ang2 = mod.delta_alpb_minus_vac(
            symbols=list(symbols),
            coords_ang=np.asarray(coords_ang, dtype=np.float64),
            charge=self.base.charge,
            multiplicity=self.base.mult,
            solvent=self.solvent,
            solvent_model=self.solvent_model,
            need_forces=need_forces,
            need_hessian=need_hessian,
            xtb_cmd=self.xtb_cmd,
            xtb_acc=self.xtb_acc,
            xtb_workdir="tmp",
            xtb_keep_files=False,
            ncores=mod.resolve_xtb_ncores(),
        )
        return float(de_ev), df_ev_ang, dh_ev_ang2

    # ------------------------------------------------------------------
    # PySisyphus API — delegate + add solvent correction
    # ------------------------------------------------------------------

    def get_energy(self, elem, coords):
        result = self.base.get_energy(elem, coords)
        if not self._enabled:
            return result
        coord_ang = np.asarray(coords, dtype=np.float64).reshape(-1, 3) * BOHR2ANG
        de_ev, _, _ = self._solvent_delta(elem, coord_ang, need_forces=False)
        result["energy"] += de_ev * _EV2AU
        return result

    def get_forces(self, elem, coords):
        result = self.base.get_forces(elem, coords)
        if not self._enabled:
            return result
        coord_ang = np.asarray(coords, dtype=np.float64).reshape(-1, 3) * BOHR2ANG
        de_ev, df_ev_ang, _ = self._solvent_delta(elem, coord_ang, need_forces=True)
        result["energy"] += de_ev * _EV2AU
        if df_ev_ang is not None:
            df_au = np.asarray(df_ev_ang, dtype=np.float64).reshape(-1) * _F_EVAA_2_AU
            result["forces"] = np.asarray(result["forces"], dtype=np.float64) + df_au
        return result

    def get_hessian(self, elem, coords):
        result = self.base.get_hessian(elem, coords)
        if not self._enabled:
            return result
        coord_ang = np.asarray(coords, dtype=np.float64).reshape(-1, 3) * BOHR2ANG
        de_ev, df_ev_ang, dh_ev_ang2 = self._solvent_delta(
            elem, coord_ang, need_forces=True, need_hessian=True
        )
        result["energy"] += de_ev * _EV2AU
        if df_ev_ang is not None:
            df_au = np.asarray(df_ev_ang, dtype=np.float64).reshape(-1) * _F_EVAA_2_AU
            result["forces"] = np.asarray(result["forces"], dtype=np.float64) + df_au
        if dh_ev_ang2 is not None:
            dh_au = np.asarray(dh_ev_ang2, dtype=np.float64) * _H_EVAA_2_AU
            n_atoms = len(elem)
            dh_au = self._apply_active_trim_np(dh_au, n_atoms)
            result["hessian"] = np.asarray(result["hessian"], dtype=np.float64) + dh_au
        return result

    # Pass through subclass hooks (unused since we override get_* directly)
    def _compute_energy_forces_ev(self, elem, coord_ang):
        return self.base._compute_energy_forces_ev(elem, coord_ang)



import itertools as it
import logging

import numpy as np
from sklearn.neighbors import KDTree

from pysisyphus.elem_data import COVALENT_RADII as CR
from pysisyphus.helpers_pure import log, timed
from pysisyphus.intcoords.valid import bend_valid, dihedral_valid


logger = logging.getLogger("internal_coords")

BOND_FACTOR = 1.3


def get_max_bond_dists(atoms, bond_factor, covalent_radii=None):
    if covalent_radii is None:
        cr = CR
    else:
        cr = {atom: covrad for atom, covrad in zip(atoms, covalent_radii)}

    unique_atoms = set(atoms)
    atom_pairs = it.combinations_with_replacement(unique_atoms, 2)
    max_bond_dists = {
        frozenset((from_, to_)): bond_factor * (cr[from_] + cr[to_])
        for from_, to_ in atom_pairs
    }
    return max_bond_dists


def find_bonds(
    atoms, coords3d, covalent_radii=None, bond_factor=BOND_FACTOR, min_dist=0.1
):
    atoms = [atom.lower() for atom in atoms]
    c3d = coords3d.reshape(-1, 3)
    if covalent_radii is None:
        covalent_radii = [CR[atom] for atom in atoms]
    cr = np.array(covalent_radii)

    max_bond_dists = get_max_bond_dists(atoms, bond_factor, covalent_radii=cr)
    radii = bond_factor * (cr.copy() + max(cr))
    kdt = KDTree(c3d)
    res, dists = kdt.query_radius(c3d, radii, return_distance=True)
    bonds_ = list()
    for i, (atom, bonds, dists_) in enumerate(zip(atoms, res, dists)):
        fsets = [frozenset((atom, atoms[a])) for a in bonds]
        ref_dists = np.array([max_bond_dists[fset] for fset in fsets])
        keep = np.logical_and(dists_ <= ref_dists, min_dist <= dists_)
        for to_ in bonds[keep]:
            bonds_.append(frozenset((i, to_)))
    bonds = np.array([tuple((from_, to_)) for from_, to_ in set(bonds_)])
    return bonds


def find_bonds_for_geom(geom, bond_factor=BOND_FACTOR):
    return find_bonds(geom.atoms, geom.coords3d, geom.covalent_radii)


def get_bond_vec_getter(
    atoms, covalent_radii, bonds_for_inds, no_bonds_with=None, bond_factor=BOND_FACTOR
):
    max_bond_dists = get_max_bond_dists(
        atoms, bond_factor, covalent_radii=covalent_radii
    )
    max_bond_dists_for_inds = [
        np.array([max_bond_dists[frozenset((atoms[ind], atom_))] for atom_ in atoms])
        for ind in bonds_for_inds
    ]
    if no_bonds_with is None:
        # List of empty lists
        no_bonds_with = [[] * len(bonds_for_inds)]

    # Set it to a negative value, so the calculated distance, which is always positive
    # can't be smaller than this value.
    for mbd, nbw in zip(max_bond_dists_for_inds, no_bonds_with):
        mbd[nbw] = -1

    all_inds = np.arange(len(atoms))

    def get_bond_vecs(coords, return_bonded_inds=False):
        coords3d = coords.reshape(-1, 3)
        all_bond_vecs = list()
        all_bonded_inds = list()
        for ind, max_dists in zip(bonds_for_inds, max_bond_dists_for_inds):
            distance_vecs = coords3d - coords3d[ind]
            distances = np.linalg.norm(distance_vecs, axis=1)
            # Set 0.0 distance of atom with itself to a high value to not form
            # and ind-ind bond.
            distances[ind] = 10_000
            bond_mask = distances <= max_dists
            all_bond_vecs.append(distance_vecs[bond_mask])
            all_bonded_inds.append(all_inds[bond_mask])

        if return_bonded_inds:
            return all_bond_vecs, all_bonded_inds
        else:
            return all_bond_vecs

    return get_bond_vecs


def get_bend_candidates(bonds):
    """Also yields duplicates [a, b, c] and [c, b, a]."""
    bond_dict = {}
    bonds = [tuple(bond) for bond in bonds]
    # Construct a dictionary holding neighbours for a given atom.
    for from_, to_ in bonds:
        bond_dict.setdefault(from_, list()).append(to_)
        bond_dict.setdefault(to_, list()).append(from_)

    for bond in bonds:
        from_, to_ = bond
        # Look up neighbours of from_ and to_ in the dictionary
        from_neighs = set(bond_dict[from_]) - set((to_,))
        to_neighs = set(bond_dict[to_]) - set((from_,))
        bend_candidates = [(neigh,) + bond for neigh in from_neighs] + [
            bond + (neigh,) for neigh in to_neighs
        ]
        yield from bend_candidates


def find_bends(coords3d, bonds, min_deg, max_deg, logger=None):
    bend_set = set()
    for indices in get_bend_candidates(bonds):
        if (
            (not bend_valid(coords3d, indices, min_deg, max_deg))
            or (indices in bend_set)
            or (indices[::-1] in bend_set)
        ):
            continue
        bend_set.add(indices)
    return [list(bend) for bend in bend_set]


def find_dihedrals(coords3d, bonds, bends, max_deg, logger=None):
    bond_dict = {}
    bonds = [tuple(bond) for bond in bonds]
    for from_, to_ in bonds:
        bond_dict.setdefault(from_, list()).append(to_)
        bond_dict.setdefault(to_, list()).append(from_)

    dihedral_set = set()
    for bend in bends:
        bend = tuple(bend)
        from_, central, to_ = bend
        from_neighs = set(bond_dict[from_]) - set((to_, central))
        to_neighs = set(bond_dict[to_]) - set((from_, central))
        dihedral_candidates = [(neigh,) + bend for neigh in from_neighs] + [
            bend + (neigh,) for neigh in to_neighs
        ]
        for indices in dihedral_candidates:
            if not dihedral_valid(coords3d, indices, deg_thresh=max_deg):
                continue
            dihedral_set.add(indices)
    return [list(dihedral) for dihedral in dihedral_set]


def find_bonds_bends(geom, bond_factor=BOND_FACTOR, min_deg=15, max_deg=175):
    log(logger, "Starting detection of bonds and bends.")
    bonds = find_bonds_for_geom(geom, bond_factor=bond_factor)
    log(logger, f"Found {len(bonds)} bonds.")
    bends = find_bends(geom.coords3d, bonds, min_deg=min_deg, max_deg=max_deg)
    log(logger, f"Found {len(bends)} bends.")
    return bonds, bends


@timed(logger)
def find_bonds_bends_dihedrals(geom, bond_factor=BOND_FACTOR, min_deg=15, max_deg=175):
    log(logger, f"Detecting bonds, bends and dihedrals for {len(geom.atoms)} atoms.")
    bonds, bends = find_bonds_bends(
        geom, bond_factor=bond_factor, min_deg=min_deg, max_deg=max_deg
    )
    proper_dihedrals = find_dihedrals(geom.coords3d, bonds, bends, max_deg)
    log(logger, f"Found {len(proper_dihedrals)} proper dihedrals.")
    return bonds, bends, proper_dihedrals



import logging

from pysisyphus import logger as pysis_logger


logger = pysis_logger.getChild("internal_coords")
logger.setLevel(logging.DEBUG)
handler = logging.FileHandler("internal_coords.log", mode="w", delay=True)
handler.setLevel(logging.DEBUG)
logger.addHandler(handler)



import argparse
import re
import sys

import numpy as np

from pysisyphus.helpers import geoms_from_trj
from pysisyphus.xyzloader import write_geoms_to_trj


def parse_args(args):
    parser = argparse.ArgumentParser()

    parser.add_argument("trj")
    parser.add_argument("filter")
    parser.add_argument("--first", type=int, metavar="N", default=None,
                        help="Only consider first N geometries in _trj.xyz."
    )

    return parser.parse_args(args)


def parse_filter(raw_filter):
    split = raw_filter.strip().split()
    mobjs = [re.match("(?P<not_present>!?)\((?P<atoms>[a-zA-Z-]+)\)", s) for s in split]
    filters = list()
    for m in mobjs:
        internal = m["atoms"]
        is_present = not bool(m["not_present"])
        internal = tuple(sorted(internal.upper().split("-")))
        filters.append((is_present, internal))
    return filters


def get_unique_internals(geom):
    attrs = ("bond_atom_indices", "bend_atom_indices", "dihedral_atom_indices")
    atoms_arr = np.array(geom.atoms)
    unique_internals = list()
    for attr in attrs:
        indices = getattr(geom.internal, attr)
        unique = set([tuple(sorted(atoms_arr[inds])) for inds in indices])
        unique_internals.extend(unique)
    return unique_internals


def run():
    args = parse_args(sys.argv[1:])
    geoms = geoms_from_trj(args.trj, first=args.first, coord_type="redund",)
    filters = parse_filter(args.filter)
    print("Filters", filters)

    valid_geoms = list()
    invalid_geoms = list()
    for i, geom in enumerate(geoms):
        internals = get_unique_internals(geom)
        valid = all([is_present == (internal in internals)
                     for is_present, internal in filters]
        )
        if not valid:
            print(f"geom {i+1} is not valid.")
            invalid_geoms.append(geom)
        else:
            valid_geoms.append(geom)
    print()
    print(f"Found {len(valid_geoms)} valid geometries.")
    print(f"Found {len(invalid_geoms)} invalid geometries.")

    write_geoms_to_trj(valid_geoms, "filtered_valid_trj.xyz")
    write_geoms_to_trj(invalid_geoms, "filtered_invalid_trj.xyz")



import numpy as np


def scale_by_max_step(steps, max_step):
    steps_max = np.abs(steps).max()
    if steps_max > max_step:
        steps *= max_step / steps_max
    return steps


def get_scale_max(max_element):
    def scale_max(step):
        step_max = np.abs(step).max()
        if step_max > max_element:
            step *= max_element / step_max
        return step
    return scale_max


def restrict_step(steps, max_step):
    too_big = np.abs(steps) > max_step
    signs = np.sign(steps[too_big])
    steps[too_big] = signs * max_step
    return steps



import logging
import random

import numpy as np


class PrimInternal:
    def __init__(self, inds, val, grad=None):
        self.inds = inds
        self.val = val
        self.grad = grad

    def __str__(self):
        return f"PrimInternal({self.inds})"

    def __repr__(self):
        return f"PrimInternal({self.inds}, {self.val:.4f})"


def eval_primitives(coords3d, primitives):
    prim_internals = list()
    for primitive in primitives:
        value, gradient = primitive.calculate(coords3d, gradient=True)
        prim_internal = PrimInternal(primitive.indices, value, gradient)
        prim_internals.append(prim_internal)
    return prim_internals


def eval_B(coords3d, primitives):
    prim_internals = eval_primitives(coords3d, primitives)
    return np.array([prim_int.grad for prim_int in prim_internals])


def check_primitives(coords3d, primitives, B=None, thresh=1e-6, logger=None):
    assert len(primitives) > 0

    def log(msg, level=logging.DEBUG):
        if logger is not None:
            logger.log(level, msg)

    if B is None:
        B = eval_B(coords3d, primitives)
    G = B.T.dot(B)
    w, v = np.linalg.eigh(G)
    nonzero_inds = np.abs(w) > thresh
    # More coordinates may be expected when collinear atoms are present.
    expected = coords3d.size - 6
    nonzero_num = sum(nonzero_inds)
    missing = expected - nonzero_num
    if missing > 0:
        log(
            "Not enough internal coordinates defined! Expected at least "
            f"{expected} nonzero eigenvalues. Found only {nonzero_num}!"
        )
    nonzero_w = w[nonzero_inds]
    # Condition number
    kappa = abs(nonzero_w.max() / nonzero_w.min())
    log(f"Condition number of B^T.B=G: {kappa:.4e}")
    return missing + 1, kappa


def augment_primitives(missing_prims, coords3d, prim_indices, fragments):
    add_bonds = list()
    add_bends = list()
    add_dihedrals = list()

    fragment_tpls = [tuple(fragment) for fragment in fragments]
    if len(fragments) > 1:
        bond_inds = prim_indices[0]
        bond_sets = [frozenset(bond) for bond in bond_inds]
        while missing_prims > 0:
            random.shuffle(fragment_tpls)
            frag1, frag2 = fragment_tpls[:2]
            atom1 = random.choice(frag1)
            atom2 = random.choice(frag2)
            bond_set = frozenset((atom1, atom2))
            if (bond_set not in bond_sets) and (bond_set not in add_bonds):
                add_bonds.append(list(bond_set))
                missing_prims -= 1
    return add_bonds, add_bends, add_dihedrals



"""In-process Hessian cache for the ``all`` workflow.

Stores Hessian matrices as CPU tensors (or numpy arrays) in a module-level
dict so that subsequent CLI stages executed via ``_run_cli_main()`` can
reuse them without recomputation.

Each entry may carry ``active_dofs`` — a list of DOF indices that the stored
(partial) Hessian spans.  Consumers set ``geometry.within_partial_hessian``
before assigning partial Hessians to ``geometry.cart_hessian``.
"""

import numpy as np
import torch
from typing import Any, Dict, Optional, Sequence

_cache: Dict[str, Any] = {}


def store(
    key: str,
    H,
    active_dofs: Optional[Sequence[int]] = None,
    meta: Optional[dict] = None,
) -> None:
    """Cache a (possibly partial) Hessian on CPU."""
    if isinstance(H, torch.Tensor):
        h_cpu = H.detach().cpu().clone()
    else:
        h_cpu = np.array(H, copy=True)
    _cache[key] = {
        "hessian": h_cpu,
        "active_dofs": list(active_dofs) if active_dofs is not None else None,
        "meta": meta or {},
    }


def load(key: str) -> Optional[Dict[str, Any]]:
    """Return cached entry or *None*."""
    return _cache.get(key)


def clear() -> None:
    """Drop all cached Hessians."""
    _cache.clear()



# tests/test_align_math.py
"""Tests for pure math utilities in pdb2reaction.align_freeze_atoms."""

import numpy as np
import pytest

from pdb2reaction.align_freeze_atoms import (
    kabsch_R_t,
    _rodrigues,
    _rotation_align_vectors,
    _orth_proj_perp,
)


class TestKabsch:
    def test_identity_case(self):
        """Same point sets should give identity rotation and zero translation."""
        P = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
        R, t = kabsch_R_t(P, P)
        np.testing.assert_allclose(R, np.eye(3), atol=1e-10)
        np.testing.assert_allclose(t, np.zeros(3), atol=1e-10)

    def test_pure_translation(self):
        """Translated point set: R = I, t = shift."""
        P = np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
        shift = np.array([3.0, -2.0, 1.0])
        Q = P - shift  # Q @ I + shift = P => t = shift
        R, t = kabsch_R_t(P, Q)
        np.testing.assert_allclose(R, np.eye(3), atol=1e-10)
        np.testing.assert_allclose(t, shift, atol=1e-10)

    def test_known_rotation(self):
        """90-degree rotation about z-axis."""
        P = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [-1.0, 0.0, 0.0]])
        # Rotate P by -90 degrees about z to get Q
        theta = -np.pi / 2
        Rz = np.array([[np.cos(theta), -np.sin(theta), 0],
                        [np.sin(theta),  np.cos(theta), 0],
                        [0, 0, 1]])
        Q = P @ Rz.T  # Row-vector rotation
        R, t = kabsch_R_t(P, Q)
        # Q @ R + t ≈ P
        reconstructed = Q @ R + t
        np.testing.assert_allclose(reconstructed, P, atol=1e-10)

    def test_reflection_handling(self):
        """Kabsch should handle reflection case (det(R) < 0 before correction)."""
        P = np.array([[1.0, 0.0, 0.0], [0.0, 2.0, 0.0], [0.0, 0.0, 3.0],
                       [1.0, 1.0, 1.0]])
        Q = P.copy()
        Q[:, 2] *= -1  # Reflect z
        R, t = kabsch_R_t(P, Q)
        assert np.linalg.det(R) > 0  # Should be proper rotation

    def test_shape_validation(self):
        with pytest.raises(ValueError, match="shape"):
            kabsch_R_t(np.zeros((3, 2)), np.zeros((3, 2)))


class TestRodrigues:
    def test_zero_angle(self):
        R = _rodrigues(np.array([0.0, 0.0, 1.0]), 0.0)
        np.testing.assert_allclose(R, np.eye(3), atol=1e-10)

    def test_90_about_z(self):
        R = _rodrigues(np.array([0.0, 0.0, 1.0]), np.pi / 2)
        expected = np.array([[0, -1, 0], [1, 0, 0], [0, 0, 1]], dtype=float)
        np.testing.assert_allclose(R, expected, atol=1e-10)

    def test_rotation_is_orthogonal(self):
        R = _rodrigues(np.array([1.0, 1.0, 1.0]), 1.23)
        np.testing.assert_allclose(R @ R.T, np.eye(3), atol=1e-10)
        np.testing.assert_allclose(np.linalg.det(R), 1.0, atol=1e-10)

    def test_zero_axis(self):
        R = _rodrigues(np.zeros(3), 1.0)
        np.testing.assert_allclose(R, np.eye(3), atol=1e-10)


class TestRotationAlignVectors:
    def test_same_direction(self):
        a = np.array([1.0, 0.0, 0.0])
        R = _rotation_align_vectors(a, a)
        np.testing.assert_allclose(R, np.eye(3), atol=1e-10)

    def test_opposite_direction(self):
        a = np.array([1.0, 0.0, 0.0])
        b = np.array([-1.0, 0.0, 0.0])
        R = _rotation_align_vectors(a, b)
        result = R @ a
        np.testing.assert_allclose(result, b, atol=1e-10)

    def test_orthogonal_vectors(self):
        a = np.array([1.0, 0.0, 0.0])
        b = np.array([0.0, 1.0, 0.0])
        R = _rotation_align_vectors(a, b)
        result = R @ a
        np.testing.assert_allclose(result, b, atol=1e-10)

    def test_arbitrary_vectors(self):
        a = np.array([1.0, 2.0, 3.0])
        b = np.array([-1.0, 0.5, 2.0])
        R = _rotation_align_vectors(a, b)
        result = R @ a
        # Should point in same direction as b
        cos_angle = np.dot(result, b) / (np.linalg.norm(result) * np.linalg.norm(b))
        np.testing.assert_allclose(cos_angle, 1.0, atol=1e-10)


class TestOrthProjPerp:
    def test_z_axis(self):
        P = _orth_proj_perp(np.array([0.0, 0.0, 1.0]))
        # Projecting z onto xy-plane should kill z component
        result = P @ np.array([1.0, 2.0, 3.0])
        np.testing.assert_allclose(result, [1.0, 2.0, 0.0], atol=1e-10)

    def test_is_idempotent(self):
        P = _orth_proj_perp(np.array([1.0, 1.0, 0.0]))
        v = np.array([1.0, 2.0, 3.0])
        np.testing.assert_allclose(P @ P @ v, P @ v, atol=1e-10)

    def test_zero_vector(self):
        P = _orth_proj_perp(np.zeros(3))
        np.testing.assert_allclose(P, np.eye(3), atol=1e-10)



import numpy as np

from pysisyphus.Geometry import Geometry
from pysisyphus.helpers import align_geoms
from pysisyphus.xyzloader import write_geoms_to_trj


class Interpolator:
    def __init__(
        self,
        geoms,
        between,
        extrapolate=0,
        extrapolate_before=0,
        extrapolate_after=0,
        extrapolate_damp=1.0,
        align=False,
    ):
        self.geoms = geoms
        self.between = between
        self.extrapolate = extrapolate
        self.extrapolate_before = (
            extrapolate_before if extrapolate_before else self.extrapolate
        )
        self.extrapolate_after = (
            extrapolate_after if extrapolate_after else self.extrapolate
        )
        self.extrapolate_damp = extrapolate_damp
        one_atom_geom = any([len(geom.atoms) == 1 for geom in geoms])
        # Don't try to align one atom species
        self.align = align and not one_atom_geom

        assert len(geoms) >= 2, "Need at least two geometries to interpolate!"

        # Check for consistent atom ordering
        for i, geom in enumerate(self.geoms[:-1]):
            next_geom = self.geoms[i + 1]
            atoms = geom.atoms
            next_atoms = next_geom.atoms
            assert len(atoms) == len(
                next_atoms
            ), f"Geometries {i} and {i+1} have a different number of atoms!"
            assert (
                atoms == next_atoms
            ), f"Different atom ordering in geometries {i} and {i+1}!"

        self.atoms = self.geoms[0].atoms
        if self.align:
            align_geoms(geoms)
        self.all_geoms = None

    def interpolate_all(self):
        all_geoms = list()

        if self.extrapolate_before:
            print("Extrapolating before")
            geoms_extrapol_before = self.interpolate(
                self.geoms[0],
                self.geoms[1],
                interpolate_only=self.extrapolate_before,
                extrapolate=True,
            )
            all_geoms += geoms_extrapol_before[::-1]

        print("Interpolating between")
        # Interpolate between all pairs of geometries
        for i, initial_geom in enumerate(self.geoms[:-1]):
            final_geom = self.geoms[i + 1]
            interpolated_geoms = self.interpolate(initial_geom, final_geom)
            assert len(interpolated_geoms) == self.between, (
                "Something is wrong with the number of interpolated "
                "geometries. Maybe you accidentally also return the "
                "initial and final geometry?"
            )
            all_geoms.append(self.geoms[i])
            all_geoms.extend(interpolated_geoms)
        # As we only added the i-th geometry and the new interpolated
        # geometries of every geometry pair we also have to add the last
        # ((i+1)-th) geometry at the end.
        all_geoms.append(self.geoms[-1])

        print("Extrapolate between")
        if self.extrapolate_after:
            geoms_extrapol_after = self.interpolate(
                self.geoms[-1],
                self.geoms[-2],
                interpolate_only=self.extrapolate_after,
                extrapolate=True,
            )
            all_geoms += geoms_extrapol_after

        self.all_geoms = all_geoms
        if self.align:
            align_geoms(self.all_geoms)
        return all_geoms

    def interpolate(
        self, initial_geom, final_geom, interpolate_only=0, extrapolate=False
    ):
        initial_coords = initial_geom.coords

        # Linear interpolation
        step = (final_geom.coords - initial_geom.coords) / (self.between + 1)
        if extrapolate:
            step *= -1
            step *= self.extrapolate_damp
        # When we extrapolate we probably want a number of geometries that is
        # different from self.between
        interpolations = interpolate_only if interpolate_only else self.between
        # initial + i*step
        i_array = np.arange(1, interpolations + 1)
        new_coords = initial_coords + i_array[:, None] * step
        return [Geometry(self.atoms, nc) for nc in new_coords]

    def all_geoms_to_trj(self, trj_fn):
        write_geoms_to_trj(self.all_geoms, trj_fn)



import struct

import numpy as np


NINE_ZEROS = struct.pack("d" * 9, *[0.0] * 9)
EYE3 = struct.pack("d" * 9, *np.eye(3).flatten())


def send_closure(sock, hdrlen, fmts, verbose=False):
    def send_msg(msg, fmt=None, packed=False):
        """Send message through socket.

        If fmt is None and the message is not packed (converted to bytes),
        it is either convert to bytes via struct with the given fmt, or it
        is assumed to be a str and encoded as ascii. Strings are padded
        to a length of hdrlen (header length, default=12).

        If it is already packed the message is sent as is, e.g. for the
        virial, which is already in bytes (packed).
        """
        if not packed:
            if fmt == "floats":
                msg = struct.pack(fmts[fmt], *msg)
            elif fmt is not None:
                msg = struct.pack(fmts[fmt], msg)
            else:
                msg = f"{msg: <{hdrlen}}".encode("ascii")
        if verbose:
            print(f"SENDING: {msg}")
        sock.sendall(msg)

    return send_msg


"""Vllt. recv_msg so verändern, dass es, wenn ein fmt gegeben ist auch
automatisch die korrekte anzahl an bytes erwartet"""


def recv_closure(sock, hdrlen, fmts, verbose=False):
    def recv_msg(nbytes=None, fmt=None, expect=""):
        """Receive a message of given length from the socket.

        If nbytes is not given we expect hdrlen (header length) bytes.
        If fmt is given the message is also unpacked using struct, otherwise
        ascii bytes are assumed and a string is returned.
        """
        if nbytes is None:
            nbytes = hdrlen

        msg = sock.recv(nbytes)
        if fmt:
            msg = struct.unpack(fmts[fmt], msg)
        else:
            msg = msg.decode("ascii").strip()
        if verbose:
            if expect:
                expect = f", expected '{expect}'"

            print(f"RECEIVED: {msg}{expect}")
        return msg

    return recv_msg


def get_fmts(cartesians):
    fmts = {
        "int": "i",  # Atom number
        "float": "d",  # Energy
        "nine_floats": "d" * 9,  # (Inverse) cell vectors, virial, zeros
        "floats": "d" * cartesians,  # Forces
        "floats_sq": "d" * cartesians**2,  # Hessian
    }
    return fmts



from math import log

from pysisyphus.config import p_DEFAULT, T_DEFAULT
from pysisyphus.constants import AU2KCALPERMOL, KBAU
from pysisyphus.helpers_pure import standard_state_corr
from pysisyphus import logger
from pysisyphus.io import geom_from_hessian


def G_aq_from_h5_hessian(h5_hessian, solv_en, T=T_DEFAULT, p=p_DEFAULT):
    geom = geom_from_hessian(h5_hessian)
    thermo = geom.get_thermoanalysis(T=T, p=p)
    dG_solv = solv_en - thermo.U_el
    G_aq = float(thermo.G + dG_solv)
    return G_aq


def direct_cycle(
    acid_h5,
    base_h5,
    acid_solv_en,
    base_solv_en,
    G_aq_H=None,
    G_gas_H=-6.28,
    dG_solv_H=-265.9,
    T=T_DEFAULT,
    p=p_DEFAULT,
):

    G_aq_acid = G_aq_from_h5_hessian(acid_h5, acid_solv_en, T=T, p=p)
    G_aq_base = G_aq_from_h5_hessian(base_h5, base_solv_en, T=T, p=p)

    if G_aq_H is None:
        G_aq_H = G_gas_H + dG_solv_H
    G_aq_H /= AU2KCALPERMOL

    dG_aq = G_aq_H + G_aq_base - G_aq_acid
    dG_aq_corr = dG_aq + standard_state_corr(T=T, p=p)

    pKa = dG_aq_corr / (KBAU * T * log(10))

    logger.debug(
        f"pKa calculation using direct cycle:\n"
        f"\tAcid HDF5: {acid_h5}\n\tBase HDF5: {base_h5}\n"
        f"\tT={T:.4f}, p={p:.1f} Pa\n"
        f"\tG_aq_acid={G_aq_acid:.6f} au\n"
        f"\tG_aq_base={G_aq_base:.6f} au\n"
        f"\tG_aq_H={G_aq_H:.6f} au\n"
        f"\tpKa={pKa:.6f}"
    )

    return pKa



import json
import socket

import numpy as np

from pysisyphus.calculators.Calculator import Calculator


class SocketCalc(Calculator):

    valid_requests = ("energy", "forces", "hessian")

    def __init__(self, *args, host="localhost", port=8080, **kwargs):
        super().__init__(*args, **kwargs)

        self.port = port
        self.host = host

    def listen_for(self, atoms, coords, request):
        request = request.lower()
        assert request.lower() in self.valid_requests, \
            f"Invalid request '{request}'! Valid requests are '{self.valid_requests}'."

        request_for = {
            "atoms": atoms,
            "coords": coords.tolist(),
            "request": request,
        }
        request_for = json.dumps(request_for).encode("utf-8")

        # Create socket
        sock = socket.socket()
        # Allow reuse
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((self.host, self.port))

        sock.listen(0)
        while True:
            client, address = sock.accept()

            with client:
                self.log(f"Got connection from {address}")

                # Send atom, coordinates and the request type
                client.sendall(request_for)

                to_json = b""
                while True:
                    data = client.recv(1024)
                    to_json += data

                    # Read until linebreak is found
                    if b"\n" in data:
                        self.log("Found linebreaking. Stop listening.")
                        break

                # Try to parse received data as JSON
                results = dict()
                try:
                    to_json.decode("utf-8")
                    results = json.loads(to_json)
                except json.JSONDecodeError:
                    self.log("JSON decode error")

                # Check if all required fields are present in the results. The
                # energy has to be always present.
                if ("energy" in results) and (request in results):
                    self.log("All required fields are present in the received JSON. "
                             "Breaking.")
                    break
                else:
                    self.log("Could not parse received data as JSON!")

        sock.close()

        # energy has to be always present
        results = {key: results[key] for key in ("energy", request)}

        if "forces" in results:
            results["forces"] = np.array(results["forces"], dtype=float)

        if "hessian" in results:
            results["hessian"] = np.array(results["hessian"],
                                          dtype=float).reshape(-1, 3*len(atoms))
        return results

    def get_energy(self, atoms, coords):
        result = self.listen_for(atoms, coords, "energy")
        return result

    def get_forces(self, atoms, coords):
        result = self.listen_for(atoms, coords, "forces")
        return result

    def get_hessian(self, atoms, coords):
        result = self.listen_for(atoms, coords, "hessian")
        return result



# [1] https://pubs.acs.org/doi/pdf/10.1021/j100180a030.
#     Toward a Systematic Molecular Orbital Theory for Excited States
#     Foresman, Head-Gordon, Pople, Frisch, 1991

import operator
from pathlib import Path
from typing import Dict, Literal, List, Optional, Tuple
import warnings

import numpy as np
from numpy.typing import NDArray

from pysisyphus.elem_data import nuc_charges_for_atoms, MASS_DICT
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers_pure import file_or_str
from pysisyphus.wavefunction.helpers import BFType
from pysisyphus.wavefunction.multipole import (
    get_multipole_moment,
    get_transition_multipole_moment,
)
from pysisyphus.wavefunction.shells import Shells


Center = Literal["coc", "com"]


class Wavefunction:
    def __init__(
        self,
        atoms: Tuple[str],
        coords: NDArray[float],
        charge: int,
        mult: int,
        unrestricted: bool,
        occ: Tuple[int],
        C: NDArray[float],
        bf_type: BFType,
        shells: Optional[Shells] = None,
        ecp_electrons=None,
        strict=True,
        warn_charge=4,
    ):
        self.atoms = atoms
        self.coords = np.array(coords).flatten()
        assert 3 * len(self.atoms) == len(self.coords)

        self.mult = mult
        if self.mult != 1:
            unrestricted = True
        self.unrestricted = unrestricted
        self.occ = occ
        if not self.unrestricted:
            assert self.occ[0] == self.occ[1]
        # MOs are stored in columns
        self.C = np.array(C)
        # Always keep α and β coefficients separate. If we get only one set
        # of MO coefficients (one square matrix with ndim == 2) we assume a
        # restricted calculation and use the same set for alpha and beta electrons.
        if C.ndim == 2:
            self.C = np.array((self.C, self.C.copy()))
        self.bf_type = bf_type
        self.shells = shells
        if ecp_electrons is None:
            ecp_electrons = np.zeros(len(self.atoms))
        elif isinstance(ecp_electrons, dict):
            _ecp_electrons = np.zeros(len(self.atoms))
            for k, v in ecp_electrons.items():
                _ecp_electrons[k] = v
            ecp_electrons = _ecp_electrons
        self.ecp_electrons = np.array(ecp_electrons, dtype=int)
        self.charge = charge
        if abs(self.charge) > warn_charge:
            warnings.warn(
                f"Encountered charge={self.charge} with high absolute value!\n"
                "This may happen in systems with ECPs. In such cases please set "
                "'ecp_electrons' or at least the correct total charge."
            )

        self._masses = np.array([MASS_DICT[atom.lower()] for atom in self.atoms])
        self._densities = dict()

        if strict:
            self.check_sanity()

    def check_sanity(self):
        assert self.shells is not None
        S = self.S  # Overlap matrix
        diag_S = np.diag(S)
        np.testing.assert_allclose(diag_S, np.ones_like(diag_S), atol=1e-12)
        Pa_mo, Pb_mo = [
            C.T @ S @ P @ S @ C for C, P in zip(self.C, self.P)
        ]  # Density matrices in MO basis
        calc_occ_a = np.trace(Pa_mo)
        calc_occ_b = np.trace(Pb_mo)
        assert np.isclose((calc_occ_a, calc_occ_b), self.occ).all()

        assert (
            self.ecp_electrons >= 0
        ).all(), "All entries in 'ecp_electrons' must be >= 0!"
        # We can't use self.nuc_charges, as it already contains ecp_electrons
        electrons_expected = (
            nuc_charges_for_atoms(self.atoms).sum()
            - self.charge  # Negative charge means MORE electrons, not less!
            - self.ecp_electrons.sum()
        )
        electrons_present = sum(self.occ)
        electrons_missing = electrons_expected - electrons_present
        assert electrons_missing == 0, (
            f"{electrons_missing} electrons are missing! Did you forget to specify "
            "'ecp_electrons' and/or the correct 'charge'?"
        )

    @property
    def atom_num(self):
        return len(self.atoms)

    @property
    def coords3d(self):
        return self.coords.reshape(-1, 3)

    @property
    def masses(self):
        return self._masses

    @property
    def nuc_charges(self):
        return nuc_charges_for_atoms(self.atoms) - self.ecp_electrons

    @property
    def mo_num(self):
        return self.C.shape[1]

    @staticmethod
    def from_file(fn, **kwargs):
        path = Path(fn)

        from_funcs = {
            ".json": Wavefunction.from_orca_json,
            ".fchk": Wavefunction.from_fchk,
        }
        from_funcs_for_str = (
            # ORCA
            ("Molden file created by orca_2mkl", Wavefunction.from_orca_molden),
            ("[AOMix Format", Wavefunction.from_aomix),
            # OpenMolcas
            # ("[N_Atoms]", Wavefunction.from_molden),  # seems buggy right now
        )
        try:
            from_func = from_funcs[path.suffix.lower()]
        except KeyError:
            # Search for certain strings in the file
            with open(fn) as handle:
                text = handle.read()
            for key, func in from_funcs_for_str:
                if key in text:
                    from_func = func
                    break
            else:
                raise Exception("Could not determine file format!")
        return from_func(fn, **kwargs)

    @staticmethod
    @file_or_str(".molden", ".molden.input")
    def from_molden(text, **kwargs):
        from pysisyphus.io.molden import wavefunction_from_molden

        wf = wavefunction_from_molden(text, **kwargs)
        return wf

    @staticmethod
    @file_or_str(".json")
    def from_orca_json(text):
        """Create wavefunction from ORCA JSON.

        As of version 5.0.3 ORCA does not create JSON files for systems
        containing an ECP, so this method does not take any additional
        args or kwargs in contrast to from_orca_molden."""
        from pysisyphus.io.orca import wavefunction_from_json

        wf = wavefunction_from_json(text)
        return wf

    @staticmethod
    @file_or_str(".molden", ".molden.input")
    def from_orca_molden(text, **kwargs):
        """Create wavefunction from ORCA molden file.

        While ORCA refuses to create JSON files for systems containing
        ECPs, this is not the case for 'orca_2mkl [fn] -molden'. So we may
        encounter ECPs here. To handle this 'wavefunction_from_molden' accepts
        an additional charge argument, to specify the correct charge, e.g. as
        stored in an ORCA calculator. If the actual, true charge is not availble
        wavefunction_from_molden will come up with an absurdly high charge.
        """

        from pysisyphus.io.orca import wavefunction_from_molden

        wf = wavefunction_from_molden(text, **kwargs)
        return wf

    @staticmethod
    @file_or_str(".molden", ".molden.input")
    def from_fchk(text, **kwargs):
        from pysisyphus.io.fchk import wavefunction_from_fchk

        wf = wavefunction_from_fchk(text, **kwargs)
        return wf

    @staticmethod
    @file_or_str(".in")
    def from_aomix(text, **kwargs):
        from pysisyphus.io.aomix import wavefunction_from_aomix

        wf = wavefunction_from_aomix(text, **kwargs)
        return wf

    @property
    def C_occ(self):
        occ_a, occ_b = self.occ
        C_a, C_b = self.C
        return C_a[:, :occ_a], C_b[:, :occ_b]

    @property
    def C_virt(self):
        occ_a, occ_b = self.occ
        C_a, C_b = self.C
        return C_a[:, occ_a:], C_b[:, occ_b:]

    @property
    def P(self):
        return [C_occ @ C_occ.T for C_occ in self.C_occ]

    @property
    def P_tot(self):
        return operator.add(*self.P)

    def P_exc(self, trans_dens):
        """
        Eqs. (2.25) and (2.26) in [1].
        """
        trans_dens *= 2**0.5 / 2
        occ_a, occ_b = self.occ
        assert occ_a == occ_b
        occ = occ_a
        occupations = np.zeros(self.mo_num)
        occupations[:occ_a] = 2
        if self.unrestricted:
            raise Exception("Fix density matrix construction!")
        P = np.diag(occupations)
        dP_oo = np.einsum("ia,ja->ij", trans_dens, trans_dens)
        dP_vv = np.einsum("ia,ic->ac", trans_dens, trans_dens)
        P[:occ, :occ] -= 2 * dP_oo
        P[occ:, occ:] += 2 * dP_vv
        C, _ = self.C
        # The density matric is currently still in the MO basis. Transform
        # it to the AO basis and return.
        return C @ P @ C.T

    def store_density(self, P, name, ao_or_mo="ao"):
        assert P.ndim == 2, "Handling of alpha/beta densities is not yet implemented!"
        if ao_or_mo == "mo":
            C, _ = self.C
            P_ao = C @ P @ C.T
        elif ao_or_mo == "ao":
            P_ao = P.copy()
        self._densities[name] = P_ao

    def get_density(self, name):
        return self._densities[name]

    def eval_density(self, coords3d, P=None):
        if P is None:
            P = self.P_tot
        spherical = self.bf_type == BFType.PURE_SPHERICAL
        vals = self.shells.eval(coords3d, spherical=spherical)
        rho = np.einsum(
            "uv,iu,iv->i", P, vals, vals, optimize=["einsum_path", (0, 1), (0, 1)]
        )
        return rho

    def eval_esp(self, coords3d, with_nuc=True):
        charges = np.ones(1)  # Assume positive charge of 1 e
        esp = np.zeros(len(coords3d))
        nuc_coords3d = self.coords3d
        Pa, Pb = self.P
        for i, c3d in enumerate(coords3d):
            V = self.get_V(c3d[None, :], charges)  # 1el Coulomb integrals
            el = np.einsum("uv,uv->", Pa, V)
            if self.unrestricted:
                el += np.einsum("uv,uv->", Pb, V)
            else:
                el *= 2.0
            if with_nuc:
                nuc = (
                    self.nuc_charges
                    / np.linalg.norm(nuc_coords3d - c3d[None, :], axis=1)
                ).sum()
            else:
                nuc = 0.0
            esp[i] = el + nuc
        return esp

    @property
    def ao_centers(self) -> List[int]:
        return list(
            {
                BFType.CARTESIAN: lambda: self.shells.cart_ao_centers,
                BFType.PURE_SPHERICAL: lambda: self.shells.sph_ao_centers,
            }[self.bf_type]()
        )

    @property
    def ao_center_map(self) -> Dict[int, List[int]]:
        ao_center_map = dict()
        for i, aoc in enumerate(self.ao_centers):
            ao_center_map.setdefault(aoc, list()).append(i)
        return ao_center_map

    def as_geom(self, **kwargs):
        return Geometry(self.atoms, self.coords, **kwargs)

    @property
    def is_cartesian(self) -> bool:
        return self.bf_type == BFType.CARTESIAN

    #####################
    # Overlap Integrals #
    #####################

    @property
    def S(self):
        # Check what type of basis functions we are using.
        return {
            BFType.CARTESIAN: lambda: self.shells.S_cart,
            BFType.PURE_SPHERICAL: lambda: self.shells.S_sph,
        }[self.bf_type]()

    def S_with_shells(self, other_shells):
        return {
            BFType.CARTESIAN: lambda: self.shells.get_S_cart(other_shells),
            BFType.PURE_SPHERICAL: lambda: self.shells.get_S_sph(other_shells),
        }[self.bf_type]()

    def S_with(self, other):
        other_shells = other.shells
        return self.S_with_shells(other_shells)

    def S_MO_with(self, other):
        assert (not self.unrestricted) and (
            self.mult == 1
        ), "Currently only Cα is considered!"
        S_AO = self.S_with(other)
        C = self.C[0]
        C_other = other.C[0]
        return C.T @ S_AO @ C_other

    #####################
    # Multipole Moments #
    #####################

    def get_origin(self, kind="coc"):
        kind = kind.lower()
        assert kind in ("com", "coc")

        coords3d = self.coords.reshape(-1, 3).copy()
        _factors = {
            "coc": self.nuc_charges,  # Center of charge
            "com": self.masses,  # Center of mass
        }
        factors = _factors[kind]
        tot_factors = factors.sum()

        return np.sum(coords3d * factors[:, None], axis=0) / tot_factors

    def dipole_ints(
        self, origin: Optional[NDArray] = None, kind: Center = "coc"
    ) -> NDArray[float]:
        if origin is None:
            origin = self.get_origin(kind=kind)

        dipole_ints = {
            BFType.CARTESIAN: lambda *args: self.shells.get_dipole_ints_cart(*args),
            BFType.PURE_SPHERICAL: lambda *args: self.shells.get_dipole_ints_sph(*args),
        }[self.bf_type](origin)
        return dipole_ints

    def quadrupole_ints(
        self, origin: Optional[NDArray] = None, kind: Center = "coc"
    ) -> NDArray[float]:
        if origin is None:
            origin = self.get_origin(kind=kind)

        quadrupole_ints = {
            BFType.CARTESIAN: lambda *args: self.shells.get_quadrupole_ints_cart(*args),
            BFType.PURE_SPHERICAL: lambda *args: self.shells.get_quadrupole_ints_sph(
                *args
            ),
        }[self.bf_type](origin)
        return quadrupole_ints

    def get_dipole_moment(
        self,
        P: Optional[NDArray[float]] = None,
        origin: Optional[NDArray[float]] = None,
        kind: Center = "coc",
    ) -> NDArray[float]:
        if origin is None:
            origin = self.get_origin(kind=kind)
        if P is None:
            P = self.P_tot
        dipole_ints = self.dipole_ints(origin)
        return get_multipole_moment(
            1, self.coords3d, origin, dipole_ints, self.nuc_charges, P
        )

    @property
    def dipole_moment(self) -> NDArray[float]:
        return self.get_dipole_moment()

    def get_quadrupole_moment(
        self,
        P: Optional[NDArray[float]] = None,
        origin: Optional[NDArray[float]] = None,
        kind: Center = "coc",
    ) -> NDArray[float]:
        if origin is None:
            origin = self.get_origin(kind=kind)
        if P is None:
            P = self.P_tot
        quadrupole_ints = self.quadrupole_ints(origin)
        return get_multipole_moment(
            2, self.coords3d, origin, quadrupole_ints, self.nuc_charges, P
        )

    @property
    def quadrupole_moment(self) -> NDArray[float]:
        return self.get_quadrupole_moment()

    ################################
    # Transition Multipole Moments #
    ################################

    def get_transition_multipole_moment(
        self,
        order: int,
        T_a: NDArray[float],
        T_b: NDArray[float] = None,
        origin: Optional[NDArray[float]] = None,
        kind: Center = "coc",
        full=False,
    ) -> NDArray[float]:
        if origin is None:
            origin = self.get_origin(kind=kind)
        if order == 1:
            multipole_ints = self.dipole_ints(origin)
        elif order == 2:
            multipole_ints = self.quadrupole_ints(origin)
        else:
            raise Exception("Multipoles of order {order} are not implemented!")
        C_a, C_b = self.C
        occ_a, occ_b = self.occ
        return get_transition_multipole_moment(
            multipole_ints,
            C_a,
            C_b,
            T_a,
            T_b,
            occ_a,
            occ_b,
            full=full,
        )

    def get_transition_dipole_moment(self, *args, **kwargs):
        return self.get_transition_multipole_moment(1, *args, **kwargs)

    def get_transition_quadrupole_moment(self, *args, **kwargs):
        return self.get_transition_multipole_moment(2, *args, **kwargs)

    def oscillator_strength(
        self, exc_ens: NDArray[float], trans_moms: NDArray[float]
    ) -> NDArray[float]:
        """Oscillator strength from TDMs and excitation energies."""
        exc_ens = np.atleast_1d(exc_ens)
        if trans_moms.ndim == 1:
            trans_moms = trans_moms[None, :]
        fosc = 2 / 3 * exc_ens * (trans_moms**2).sum(axis=1)
        return fosc

    #########################
    # 1el Coulomb Integrals #
    #########################

    def get_V(self, coords3d, charges):
        return {
            BFType.CARTESIAN: self.shells.get_V_cart,
            BFType.PURE_SPHERICAL: self.shells.get_V_sph,
        }[self.bf_type](coords3d, charges)

    def __str__(self):
        is_restricted = "unrestricted" if self.unrestricted else "restricted"
        return (
            f"Wavefunction({self.atom_num} atoms, charge={self.charge}, {is_restricted}, "
            f"mult={self.mult})"
        )



# pdb2reaction/backends/base.py

"""
Base MLIP calculator class for PySisyphus integration.

All MLIP backends (UMA, ORB, MACE, AIMNet2) extend ``MLIPCalculator``,
which inherits from ``pysisyphus.calculators.Calculator``.  Common
functionality — freeze-atom handling, finite-difference Hessian assembly,
unit conversion, and timing/VRAM logging — lives here.
"""

from __future__ import annotations

import time
from typing import Any, Dict, List, Optional, Sequence

import click
import numpy as np

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.constants import BOHR2ANG, ANG2BOHR, AU2EV

# ---- Unit conversion constants ----
EV2AU = 1.0 / AU2EV                       # eV → Hartree
F_EVAA_2_AU = EV2AU / ANG2BOHR            # eV Å⁻¹ → Hartree Bohr⁻¹
H_EVAA_2_AU = EV2AU / ANG2BOHR / ANG2BOHR # eV Å⁻² → Hartree Bohr⁻²


class BackendError(RuntimeError):
    """Raised for backend-specific runtime failures."""


class MLIPCalculator(Calculator):
    """PySisyphus-compatible MLIP calculator base class.

    Subclasses must implement:

    * ``_compute_energy_forces_ev(elem, coord_ang)``
      → ``(energy_eV, forces_eV_Ang)``
    * optionally ``_compute_analytical_hessian_ev(elem, coord_ang)``
      → ``hessian_eV_Ang2`` (3N×3N ndarray or torch Tensor)

    The base class provides ``get_energy``, ``get_forces``, ``get_hessian``
    methods that handle freeze-atom masking, unit conversion, and FD Hessian
    assembly automatically.
    """

    implemented_properties = ["energy", "forces", "hessian"]

    def __init__(
        self,
        *,
        charge: int = 0,
        spin: int = 1,
        device: str = "auto",
        freeze_atoms: Optional[Sequence[int]] = None,
        hessian_calc_mode: str = "FiniteDifference",
        return_partial_hessian: bool = False,
        hessian_double: bool = True,
        print_timing: bool = True,
        **kwargs,
    ):
        super().__init__(charge=charge, mult=spin, **kwargs)

        self.device_str = device
        self.hessian_calc_mode = hessian_calc_mode
        self.return_partial_hessian = bool(return_partial_hessian)
        self.hessian_double = bool(hessian_double)
        self.print_timing = bool(print_timing)

        # Normalise freeze_atoms
        if freeze_atoms is None:
            freeze_iter: List[int] = []
        else:
            try:
                freeze_iter = [int(i) for i in list(freeze_atoms)]
            except Exception:
                freeze_iter = []
        self.freeze_atoms = sorted(set(freeze_iter))

    # ------------------------------------------------------------------
    # Subclass hooks (override in backend implementations)
    # ------------------------------------------------------------------

    def _compute_energy_forces_ev(
        self, elem: Sequence[str], coord_ang: np.ndarray
    ) -> tuple:
        """Return ``(energy_eV: float, forces_eV_Ang: ndarray(N,3))``."""
        raise NotImplementedError

    def _compute_analytical_hessian_ev(
        self, elem: Sequence[str], coord_ang: np.ndarray
    ) -> Any:
        """Return Hessian in eV/Å² as (3N,3N) ndarray or torch Tensor.

        Raise ``NotImplementedError`` if analytical Hessian is not available
        for this backend.  The base class will fall back to FD.
        """
        raise NotImplementedError(
            f"Analytical Hessian is not available for {self.__class__.__name__}."
        )

    def _supports_analytical_hessian(self) -> bool:
        """Return True if this backend can compute analytical Hessians."""
        return False

    # ------------------------------------------------------------------
    # Freeze-atom helpers
    # ------------------------------------------------------------------

    def _active_and_frozen_dof_idx(self, n_atoms: int):
        frozen_set = set(self.freeze_atoms)
        active_atoms = [i for i in range(n_atoms) if i not in frozen_set]
        active_dof_idx = [3 * i + j for i in active_atoms for j in range(3)]
        frozen_dof_idx = [3 * i + j for i in self.freeze_atoms for j in range(3)]
        return active_atoms, active_dof_idx, frozen_dof_idx

    def _build_partial_hessian_meta(self, n_atoms: int) -> Optional[dict]:
        """Return the partial-Hessian metadata dict (to be stored as
        ``out['within_partial_hessian']``) when partial Hessian is active,
        otherwise ``None``."""
        if not self.return_partial_hessian or len(self.freeze_atoms) == 0:
            return None
        active_atoms, active_dof_idx, _ = self._active_and_frozen_dof_idx(n_atoms)
        return {
            "active_atoms": np.array(active_atoms, dtype=int),
            "active_dofs": np.array(active_dof_idx, dtype=int),
            "active_n_dof": len(active_dof_idx),
            "full_n_dof": n_atoms * 3,
        }

    def _zero_frozen_forces_ev(self, F: np.ndarray) -> np.ndarray:
        """Zero forces (eV/Å) on frozen atoms."""
        if (F is None) or (len(self.freeze_atoms) == 0):
            return F
        Fz = F.copy()
        Fz[np.asarray(self.freeze_atoms, dtype=int)] = 0.0
        return Fz

    # ------------------------------------------------------------------
    # FD Hessian (CPU, used by non-UMA backends)
    # ------------------------------------------------------------------

    def _build_fd_hessian_cpu(
        self,
        elem: Sequence[str],
        coord_ang: np.ndarray,
        *,
        eps_ang: float = 1.0e-3,
    ) -> Dict[str, Any]:
        """Central-difference Hessian assembled on CPU.

        Frozen-atom DOFs are never displaced, so their rows and columns remain
        zero in the full (3N×3N) output. When ``return_partial_hessian=True``
        the active-DOF sub-block is extracted and returned instead.
        """
        n_atoms = len(elem)
        dof = n_atoms * 3

        frozen_set = set(self.freeze_atoms)
        active_atoms = [i for i in range(n_atoms) if i not in frozen_set]
        active_dof_idx = [3 * i + j for i in active_atoms for j in range(3)]

        # Base point
        e0, F0 = self._compute_energy_forces_ev(elem, coord_ang)
        F0 = np.asarray(F0, dtype=np.float64).reshape(n_atoms, 3)

        H = np.zeros((dof, dof), dtype=np.float64)
        coord_plus = coord_ang.copy()
        coord_minus = coord_ang.copy()

        for k in active_dof_idx:
            a = k // 3
            c = k % 3

            coord_plus[a, c] = coord_ang[a, c] + eps_ang
            _, Fp = self._compute_energy_forces_ev(elem, coord_plus)
            Fp = np.asarray(Fp, dtype=np.float64).reshape(-1)

            coord_minus[a, c] = coord_ang[a, c] - eps_ang
            _, Fm = self._compute_energy_forces_ev(elem, coord_minus)
            Fm = np.asarray(Fm, dtype=np.float64).reshape(-1)

            H[:, k] = -(Fp - Fm) / (2.0 * eps_ang)

            coord_plus[a, c] = coord_ang[a, c]
            coord_minus[a, c] = coord_ang[a, c]

        # Symmetrise
        H = 0.5 * (H + H.T)

        # Reduce to active-DOF block if requested
        if self.return_partial_hessian:
            idx = np.array(active_dof_idx, dtype=int)
            H = H[np.ix_(idx, idx)]

        return {"energy": e0, "forces": F0, "hessian": H}

    # ------------------------------------------------------------------
    # Hessian postprocessing
    # ------------------------------------------------------------------

    def _hessian_ev_to_au(self, H_ev_ang2: np.ndarray) -> np.ndarray:
        """Convert (3N,3N) Hessian from eV/Å² to Hartree/Bohr² and symmetrise."""
        H = np.asarray(H_ev_ang2, dtype=np.float64)
        if H.ndim == 4:
            n = H.shape[0]
            H = H.reshape(n * 3, n * 3)
        H = 0.5 * (H + H.T)
        H *= H_EVAA_2_AU
        return H

    def _apply_active_trim_np(self, H: np.ndarray, n_atoms: int) -> np.ndarray:
        """Extract the active-DOF sub-block when ``return_partial_hessian=True``,
        or zero frozen rows/columns when ``False`` (numpy version)."""
        if len(self.freeze_atoms) == 0:
            return H
        _, active_dof_idx, frozen_dof_idx = self._active_and_frozen_dof_idx(n_atoms)
        dof = n_atoms * 3
        H = H.reshape(dof, dof)
        if self.return_partial_hessian:
            idx = np.array(active_dof_idx, dtype=int)
            return H[np.ix_(idx, idx)]
        else:
            if frozen_dof_idx:
                H[:, frozen_dof_idx] = 0.0
                H[frozen_dof_idx, :] = 0.0
            return H

    # ------------------------------------------------------------------
    # PySisyphus API
    # ------------------------------------------------------------------

    def get_energy(self, elem, coords):
        coord_ang = np.asarray(coords, dtype=np.float64).reshape(-1, 3) * BOHR2ANG
        e_ev, _ = self._compute_energy_forces_ev(elem, coord_ang)
        return {"energy": e_ev * EV2AU}

    def get_forces(self, elem, coords):
        coord_ang = np.asarray(coords, dtype=np.float64).reshape(-1, 3) * BOHR2ANG
        e_ev, F_ev = self._compute_energy_forces_ev(elem, coord_ang)
        F_ev = self._zero_frozen_forces_ev(np.asarray(F_ev, dtype=np.float64).reshape(-1, 3))
        return {
            "energy": e_ev * EV2AU,
            "forces": (F_ev.reshape(-1) * F_EVAA_2_AU),
        }

    def get_hessian(self, elem, coords):
        coord_ang = np.asarray(coords, dtype=np.float64).reshape(-1, 3) * BOHR2ANG
        n_atoms = coord_ang.shape[0]

        hess_total_start = time.perf_counter()
        mode_label = "FiniteDifference"

        mode = (self.hessian_calc_mode or "FiniteDifference").strip().lower()
        use_analytical = mode in ("analytical", "analytic")

        if use_analytical and self._supports_analytical_hessian():
            mode_label = "Analytical"
            t0 = time.perf_counter()
            e_ev, F_ev = self._compute_energy_forces_ev(elem, coord_ang)
            H_ev = self._compute_analytical_hessian_ev(elem, coord_ang)
            H_ev = np.asarray(H_ev, dtype=np.float64)
            if H_ev.ndim == 4:
                H_ev = H_ev.reshape(n_atoms * 3, n_atoms * 3)
            mode_elapsed = time.perf_counter() - t0

            F_ev = self._zero_frozen_forces_ev(
                np.asarray(F_ev, dtype=np.float64).reshape(-1, 3)
            )
            H_ev = self._apply_active_trim_np(H_ev, n_atoms)
            H_au = self._hessian_ev_to_au(H_ev)

            out = {
                "energy": e_ev * EV2AU,
                "forces": (F_ev.reshape(-1) * F_EVAA_2_AU),
                "hessian": H_au,
            }
        else:
            t0 = time.perf_counter()
            res = self._build_fd_hessian_cpu(elem, coord_ang)
            mode_elapsed = time.perf_counter() - t0

            F_ev = self._zero_frozen_forces_ev(
                np.asarray(res["forces"], dtype=np.float64).reshape(-1, 3)
            )
            H_au = self._hessian_ev_to_au(res["hessian"])

            out = {
                "energy": res["energy"] * EV2AU,
                "forces": (F_ev.reshape(-1) * F_EVAA_2_AU),
                "hessian": H_au,
            }

        total_elapsed = time.perf_counter() - hess_total_start
        if self.print_timing:
            click.echo(f"[HessianTiming] mode: {mode_label} | elapsed: {mode_elapsed:.2f} s")

        partial_meta = self._build_partial_hessian_meta(n_atoms)
        if partial_meta is not None:
            out["within_partial_hessian"] = partial_meta

        return out



from pysisyphus.line_searches.LineSearch import (
    LineSearch,
    LineSearchNotConverged,
)

from pysisyphus.line_searches.interpol import interpol_alpha_quad, interpol_alpha_cubic
from pysisyphus.optimizers.poly_fit import cubic_fit, quartic_fit


class Backtracking(LineSearch):
    def __init__(self, *args, rho_lo=5e-2, rho_hi=0.9, use_grad=False, **kwargs):
        """Backtracking line search enforcing Armijo conditions.

        Uses only energy evaluations.

        See [1], Chapter 3, Line Search methods, Section 3.1 p. 31 and
        Section 3.5 p. 56."""

        kwargs["cond"] = "armijo"
        super().__init__(*args, **kwargs)

        self.rho_lo = float(rho_lo)
        self.rho_hi = float(rho_hi)
        self.use_grad = use_grad

    def alpha_new_from_phi(self, cycle, phi0, dphi0, alpha, alpha_prev):
        phi_i = self.get_phi_dphi("f", alpha)
        self.log(f"\tCycle {cycle:02d}: alpha={alpha:.6f}, ϕ={phi_i:.6f} au")

        if cycle == 0:
            # Quadratic interpolation
            alpha_new = interpol_alpha_quad(phi0, dphi0, phi_i, alpha)
            type_ = "quadratic"
        else:
            # Cubic interpolation
            phi_prev = self.get_phi_dphi("f", alpha_prev)
            alpha_new = interpol_alpha_cubic(
                phi0, dphi0, phi_prev, phi_i, alpha_prev, alpha
            )
            type_ = "cubic"
        return alpha_new, type_

    def alpha_new_from_phi_dphi(self, cycle, phi0, dphi0, alpha):
        phi_i, dphi_i = self.get_phi_dphi("fg", alpha)
        self.log(f"\tCycle {cycle:02d}: α={alpha:.6f}, ϕ={phi_i:.6f} au")

        # First we try a constrained quartic polynomial
        res = quartic_fit(phi0, phi_i, dphi0, dphi_i)
        type_ = "quartic"
        # If the quartic poly failed, we continue with a cubic polynomial
        if res is None:
            res = cubic_fit(phi0, phi_i, dphi0, dphi_i)
            type_ = "cubic"
        # If the cubic poly failed we resort to bisection. ohoh
        if res is None:
            alpha_new = 0.5 * alpha
            type_ = "bisection"
        else:
            alpha_new = res.x * alpha
        return alpha_new, type_

    def run_line_search(self):
        phi0, dphi0 = self.get_phi_dphi("fg", 0)

        alpha_prev = None
        alpha = self.alpha_init
        for i in range(self.max_cycles):
            if self.use_grad:
                alpha_new, type_ = self.alpha_new_from_phi_dphi(i, phi0, dphi0, alpha)
            else:
                alpha_new, type_ = self.alpha_new_from_phi(i, phi0, dphi0, alpha, alpha_prev)
            self.log(f"\tNew α from {type_} interpolation: {alpha_new:.6f}")

            lower_bound = alpha * self.rho_lo
            upper_bound = alpha * self.rho_hi
            if alpha_new < lower_bound:
                self.log("\tNew α is too small!")
            if alpha_new > upper_bound:
                self.log("\tNew α is too big!")

            # Assert that alpha doesn't change too much compared to the previous alpha
            alpha_new = min(alpha_new, upper_bound)
            alpha_new = max(alpha_new, lower_bound)
            alpha_prev = alpha
            alpha = alpha_new
            self.log(f"\tNext α: {alpha:.6f}\n")

        raise LineSearchNotConverged



import json
import urllib.request
from urllib.error import HTTPError

from pysisyphus.io.sdf import geom_from_sdf


def cid_from_name(name):
    url = f"https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/name/{name}/property/IUPACName/JSON"
    with urllib.request.urlopen(url) as handle:
        text = handle.read()
    json_ = json.loads(text)
    cid = json_["PropertyTable"]["Properties"][0]["CID"]
    return cid


def sdf_from_cid(cid):
    url = f"https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/cid/{cid}/SDF?record_type=3d"
    with urllib.request.urlopen(url) as handle:
        sdf = handle.read().decode("utf-8")
    return sdf


def geom_from_pubchem_name(name, **kwargs):
    try:
        cid = cid_from_name(name)
        sdf = sdf_from_cid(cid)
        geom = geom_from_sdf(sdf, **kwargs)
    except HTTPError:
        geom = None
    return geom



#!/usr/bin/env python3
"""Smoke test core CLI subcommands with lightweight fixtures."""

from __future__ import annotations

import argparse
import tempfile
from pathlib import Path

from click.testing import CliRunner

REPO_ROOT = Path(__file__).resolve().parents[1]
FIXTURE_ROOT = REPO_ROOT / "tests" / "smoke"


def _fixture_paths() -> dict[str, Path]:
    paths = {
        "r_pdb": FIXTURE_ROOT / "r.pdb",
        "p_pdb": FIXTURE_ROOT / "p.pdb",
        "r_complex_pdb": FIXTURE_ROOT / "r_complex.pdb",
        "ts_gjf": FIXTURE_ROOT / "ts.gjf",
    }
    missing = [str(path) for path in paths.values() if not path.exists()]
    if missing:
        raise RuntimeError(
            "[core-cli-smoke] missing required fixtures:\n  - " + "\n  - ".join(missing)
        )
    return {k: v.resolve() for k, v in paths.items()}


def _invoke_or_raise(runner: CliRunner, args: list[str], label: str) -> None:
    from pdb2reaction.cli import cli as root_cli

    result = runner.invoke(root_cli, args, catch_exceptions=False)
    if result.exit_code != 0:
        raise RuntimeError(
            f"[core-cli-smoke] failed: {label}\n"
            f"command: pdb2reaction {' '.join(args)}\n"
            f"exit_code: {result.exit_code}\n"
            f"output:\n{result.output}"
        )


def _run_help_smoke(runner: CliRunner) -> int:
    subcommands = ["extract", "path-opt", "path-search", "tsopt", "freq", "irc", "dft"]
    for subcmd in subcommands:
        _invoke_or_raise(runner, [subcmd, "--help"], f"{subcmd} --help")
        _invoke_or_raise(runner, [subcmd, "--help-advanced"], f"{subcmd} --help-advanced")
    return len(subcommands)


def _run_dry_run_smoke(runner: CliRunner, fixtures: dict[str, Path]) -> int:
    commands: list[list[str]] = [
        [
            "path-opt",
            "-i",
            str(fixtures["r_pdb"]),
            str(fixtures["p_pdb"]),
            "-q",
            "0",
            "--dry-run",
        ],
        [
            "path-search",
            "-i",
            str(fixtures["r_pdb"]),
            "-i",
            str(fixtures["p_pdb"]),
            "-q",
            "0",
            "--dry-run",
        ],
        ["tsopt", "-i", str(fixtures["ts_gjf"]), "--dry-run"],
        ["freq", "-i", str(fixtures["ts_gjf"]), "--dry-run"],
        ["irc", "-i", str(fixtures["ts_gjf"]), "--dry-run"],
        ["dft", "-i", str(fixtures["ts_gjf"]), "--dry-run"],
    ]
    for cmd in commands:
        _invoke_or_raise(runner, cmd, " ".join(cmd[:1]) + " --dry-run")
    return len(commands)


def _run_extract_smoke(runner: CliRunner, fixtures: dict[str, Path]) -> None:
    with tempfile.TemporaryDirectory(prefix="pdb2reaction_extract_smoke_") as tmpdir:
        output_pdb = Path(tmpdir) / "pocket.pdb"
        _invoke_or_raise(
            runner,
            [
                "extract",
                "-i",
                str(fixtures["r_complex_pdb"]),
                "-c",
                "PRE",
                "-o",
                str(output_pdb),
                "--no-verbose",
            ],
            "extract execution smoke",
        )
        if not output_pdb.exists():
            raise RuntimeError(
                "[core-cli-smoke] extract succeeded but output pocket was not created: "
                f"{output_pdb}"
            )


def main() -> int:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.parse_args()

    fixtures = _fixture_paths()
    runner = CliRunner()

    n_help = _run_help_smoke(runner)
    n_dry = _run_dry_run_smoke(runner, fixtures)
    _run_extract_smoke(runner, fixtures)

    print(f"[core-cli-smoke] help validated for {n_help} subcommands.")
    print(f"[core-cli-smoke] dry-run validated for {n_dry} subcommands.")
    print("[core-cli-smoke] extract execution smoke passed.")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from dataclasses import dataclass
from enum import Enum
import logging
import os
from pathlib import Path
import platform
import shutil
import subprocess
import tempfile
from typing import Callable, Optional

from natsort import natsorted

from pysisyphus import logger
from pysisyphus import helpers_pure
from pysisyphus.config import get_cmd, OUT_DIR_DEFAULT
from pysisyphus.constants import BOHR2ANG
from pysisyphus.helpers import geom_loader
from pysisyphus.linalg import finite_difference_hessian


KeepKind = Enum("KeepKind", ["ALL", "LATEST", "NONE"])
HessKind = Enum("HessKind", ["ORG", "NUMERICAL"])


@dataclass
class SetPlan:
    key: str
    name: Optional[str] = None
    condition: Callable = lambda obj: True
    # success: Optional[Callable] = None
    fail: Optional[Callable] = None

    def __post_init__(self):
        if self.name is None:
            self.name = self.key


class Calculator:

    conf_key = None
    _set_plans = []

    def __init__(
        self,
        calc_number=0,
        charge=0,
        mult=1,
        base_name="calculator",
        pal=1,
        mem=1000,
        keep_kind="all",
        check_mem=True,
        retry_calc=0,
        last_calc_cycle=None,
        clean_after=True,
        out_dir=OUT_DIR_DEFAULT,
        force_num_hess=False,
        num_hess_kwargs=None,
    ):
        """Base-class of all calculators.

        Meant to be extended.

        Parameters
        ----------
        calc_number : int, default=0
            Identifier of the Calculator. Used in distinguishing it from
            other Calculators, e.g. in ChainOfStates calculations. Also
            used in the creation of filenames.
        charge : int, default=0
            Molecular charge.
        mult : int, default=1
            Molecular multiplicity (1 = singlet, 2 = doublet, ...)
        base_name : str, default=calculator
            Generated filenames will start with this string.
        pal : int, default=1
            Positive integer that gives the number of physical cores to
            use on 1 node.
        mem : int, default=1000
            Mememory per core in MB. The total amount of memory is given as
            mem*pal.
        check_mem : bool, default=True
            Whether to adjust the requested memory if too much is requested.
        retry_calc : int, default=0
            Number of additional retries when calculation failed.
        last_calc_cycle : int
            Internal variable used in restarts.
        clean_after : bool
            Delete temporary directory were calculations were executed
            after a calculation.
        out_dir : str
            Path that is prepended to generated filenames.
        force_hess_kwargs : bool, default False
            Force numerical Hessians.
        num_hess_kwargs : dict
            Keyword arguments for finite difference Hessian calculation.
        """

        self.logger = logging.getLogger("calculator")

        self.calc_number = calc_number
        self.charge = int(charge)
        self.mult = int(mult)
        self.base_name = base_name
        self.pal = int(pal)
        assert self.pal > 0, "pal must be a non-negative integer!"
        if check_mem:
            mem = helpers_pure.check_mem(int(mem), pal, logger=self.logger)
        self.mem = mem
        self.keep_kind = KeepKind[keep_kind.upper()]
        # Disasble retries if check_termination method is not implemented
        self.retry_calc = int(retry_calc) if hasattr(self, "check_termination") else 0
        assert self.retry_calc >= 0
        try:
            self.out_dir = Path(out_dir).resolve()
        except TypeError:
            self.out_dir = Path(OUT_DIR_DEFAULT).resolve()
        self.out_dir.mkdir(parents=True, exist_ok=True)
        if num_hess_kwargs is None:
            num_hess_kwargs = dict()
        self.num_hess_kwargs = num_hess_kwargs

        # Extensions of the files to keep after running a calculation.
        # Usually overridden in derived classes.
        self.to_keep = ()
        # How many calculations were already run
        self.calc_counter = 0
        # Handle restarts
        if last_calc_cycle:
            self.calc_counter = int(last_calc_cycle) + 1
            self.reattach(int(last_calc_cycle))
            self.log(f"Set {self.calc_counter} for this calculation")
        self.clean_after = clean_after

        self.inp_fn = "calc.inp"
        self.out_fn = "calc.out"
        # When this is set the run() method will use this path
        # instead of creating a new one.
        # Currently this is only used with the Turbomole calculator.
        self.path_already_prepared = None
        self.last_run_path = None
        self.backup_dir = None
        self.kept_history = dict()
        self.build_set_plans()

        # Backup original get_hessian method
        self._org_get_hessian = self.get_hessian
        self.hessian_kind = HessKind["ORG"]
        if force_num_hess:
            self.force_num_hessian()

    def get_cmd(self, key="cmd"):
        assert self.conf_key, "'conf_key'-attribute is missing for this calculator!"

        try:
            return get_cmd(section=self.conf_key, key=key, use_defaults=True)
        except KeyError:
            logger.debug(f"Failed to load key '{key}' from section '{self.conf_key}'!")

    @classmethod
    def geom_from_fn(cls, fn, **kwargs):
        geom = geom_loader(fn)
        calc = cls(**kwargs)
        geom.set_calculator(calc)
        return geom

    @property
    def name(self):
        return f"{self.base_name}_{self.calc_number:03d}"

    def log(self, message=""):
        """Write a log message.

        Wraps the logger variable.

        Parameters
        ----------
        message : str
            Message to be logged.
        """

        self.logger.debug(f"{self.name}, cycle {self.calc_counter:03d}: {message}")

    def get_energy(self, atoms, coords, **prepare_kwargs):
        """Meant to be extended."""
        raise Exception("Not implemented!")

    def get_forces(self, atoms, coords, **prepare_kwargs):
        """Meant to be extended."""
        raise Exception("Not implemented!")

    def get_hessian(self, atoms, coords, **prepare_kwargs):
        """Get Hessian matrix. Fall back to numerical Hessian, if not overriden.

        Preferrably, this method should provide an analytical Hessian."""
        self.log(f"Calculator {self} has no native Hessian method.")
        return self.get_num_hessian(atoms, coords, **prepare_kwargs)

    def get_num_hessian(self, atoms, coords, **prepare_kwargs):
        self.log("Calculating numerical Hessian.")
        results = self.get_energy(atoms, coords, **prepare_kwargs)

        def grad_func(coords):
            results = self.get_forces(atoms, coords, **prepare_kwargs)
            gradient = -results["forces"]
            return gradient

        def callback(i, j):
            self.log(f"Displacement {j} of coordinate  {i}")

        _num_hess_kwargs = {
            "step_size": 0.005,
            # Central difference by default
            "acc": 2,
        }
        _num_hess_kwargs.update(self.num_hess_kwargs)

        fd_hessian = finite_difference_hessian(
            coords,
            grad_func,
            callback=callback,
            **_num_hess_kwargs,
        )
        results["hessian"] = fd_hessian
        return results

    def force_num_hessian(self):
        """Always calculate numerical Hessians."""
        self.log("Doing numerical Hessians from now on.")
        self.get_hessian = self.get_num_hessian
        self.hessian_kind = HessKind["NUMERICAL"]

    def restore_org_hessian(self):
        """Restore original 'get_hessian' method, which may also fallback to numerical
        Hessians, if not implemented."""
        self.log("Restored original 'get_hessian' method.")
        self.get_hessian = self._org_get_hessian
        self.hessian_kind = HessKind["ORG"]

    def make_fn(self, name, counter=None, return_str=False):
        """Make a full filename.

        Return a full filename including the calculator name and the
        current counter given a suffix.

        Parameters
        ----------
        name: str
            Suffix of the filename.
        counter : int, optional
            If not given use the current calc_counter.
        return_str : int, optional
            Return a string instead of a Path when True.

        Returns
        -------
        fn : str
            Filename.
        """

        if counter is None:
            counter = self.calc_counter
        fn = self.out_dir / f"{self.name}.{counter:03d}.{name}"
        if return_str:
            fn = str(fn)
        return fn

    def prepare_path(self, use_in_run=False):
        """Get a temporary directory handle.

        Create a temporary directory that can later be used in a calculation.

        Parameters
        ----------
            use_in_run : bool, option
                Sets the internal variable ``self.path_already_prepared`` that
                is later read by ``self.run()``. No new temporary directory will
                be created in ``self.run()``.

        Returns
        -------
            path: Path
                Prepared directory.
        """

        prefix = f"{self.name}_{self.calc_counter:03d}_"
        path = Path(tempfile.mkdtemp(prefix=prefix))
        if use_in_run:
            self.path_already_prepared = path
        return path

    def prepare(self, inp):
        """Prepare a temporary directory and write input.

        Similar to prepare_path, but the input is also written into
        the prepared directory.

        Paramters
        ---------
        inp : str
            Input to be written into the file ``self.inp_fn`` in
            the prepared directory.

        Returns
        -------
            path: Path
                Prepared directory.
        """

        if not self.path_already_prepared:
            path = self.prepare_path()
        else:
            path = self.path_already_prepared

        # Calculators like Turbomole got no input.
        if inp:
            inp_path = path / self.inp_fn
            with open(inp_path, "w") as handle:
                handle.write(inp)

        return path

    def prepare_input(self, atoms, coords, calc_type):
        """Meant to be extended."""
        raise Exception("Not implemented!")

    def print_out_fn(self, path):
        """Print calculation output.

        Prints the output of a calculator after a calculation.

        Parameters
        ----------
        path : Path
            Temporary directory of the calculation.
        """
        with open(path / self.out_fn) as handle:
            text = handle.read()
        print(text)

    def prepare_turbo_coords(self, atoms, coords):
        """Get a Turbomole coords string.

        Parameters
        ----------
        atoms : iterable
            Atom descriptors (element symbols).
        coords: np.array, 1d
            1D-array holding coordinates in Bohr.

        Returns
        -------
        coords: str
            String holding coordinates in Turbomole coords format.
        """
        fmt = "{:<20.014f}"
        coord_str = "$coord\n"
        for atom, coord in zip(atoms, coords.reshape(-1, 3)):
            coord_line = (fmt + fmt + fmt).format(*coord) + atom.lower() + "\n"
            coord_str += coord_line
        coord_str += "$end"
        return coord_str

    def prepare_coords(self, atoms, coords):
        """Get 3d coords in Angstrom.

        Reshape internal 1d coords to 3d and convert to Angstrom.

        Parameters
        ----------
        atoms : iterable
            Atom descriptors (element symbols).
        coords: np.array, 1d
            1D-array holding coordinates in Bohr.

        Returns
        -------
        coords: np.array, 3d
            3D-array holding coordinates in Angstrom.
        """
        coords = coords.reshape(-1, 3) * BOHR2ANG
        coords = "\n".join(
            [
                "{} {:10.08f} {:10.08f} {:10.08f}".format(a, *c)
                for a, c in zip(atoms, coords)
            ]
        )
        return coords

    def prepare_xyz_string(self, atoms, coords):
        """Returns a xyz string in Angstrom.

        Parameters
        ----------
        atoms : iterable
            Atom descriptors (element symbols).
        coords: np.array, 1d
            1D-array holding coordinates in Bohr.

        Returns
        -------
        xyz_str: string
            Coordinates in .xyz format.
        """

        return f"{len(atoms)}\n\n{self.prepare_coords(atoms, coords)}"

    def run(
        self,
        inp,
        calc,
        add_args=None,
        env=None,
        shell=False,
        hold=False,
        keep=True,
        cmd=None,
        inc_counter=True,
        run_after=True,
        parser_kwargs=None,
        symlink=True,
    ):
        """Run a calculation.

        The bread-and-butter method to actually run an external quantum
        chemistry code.

        Parameters
        ----------
        inp : str
            Input for the external program that is written to the temp-dir.
        calc : str, hashable
            Key (and more or less type of calculation) to select the right
            parsing function from ``self.parser_funcs``.
        add_args : iterable, optional
            Additional arguments that will be appended to the program call.
        env : Environment, optional
            A potentially modified environment for the subprocess call.
        shell : bool, optional
            Use a shell to execute the program call. Need for Turbomole were
            we chain program calls like dscf; escf.
        hold : bool, optional
            Wether to remove the temporary directory after the calculation.
        keep : bool, optional
            Wether to backup files as specified in ``self.to_keep()``. Usually
            you want this.
        cmd : str or iterable, optional
            Overwrites ``self.base_cmd``.
        inc_counter : bool, optional
            Wether to increment the counter after a calculation.

        Returns
        -------
        results : dict
            Dictionary holding all applicable results of the calculations
            like the energy, a forces vector and/or excited state energies
            from TDDFT.
        """

        self.backup_dir = None
        path = self.prepare(inp)
        self.log(f"Running in {path} on {platform.node()}")
        if cmd is None:
            cmd = self.base_cmd

        if isinstance(cmd, str):
            cmd = [cmd]

        act_cmd = cmd[0]
        cmd_availble = shutil.which(act_cmd)
        if (not cmd_availble) and (not shell):
            print(
                f"Command '{act_cmd}' could not be found on $PATH! "
                "Maybe you forgot to make it available?"
            )
            return

        args = cmd + [self.inp_fn]
        if add_args:
            args.extend(add_args)
        if not env:
            env = os.environ.copy()
        tmp_out_fn = path / self.out_fn
        with open(tmp_out_fn, "w") as handle:
            if symlink:
                # We can't use resolve here as a previous symlink may already
                # exist. Calling resolve would translate this to the original
                # out file in some tempdir (that is already deleted ...).
                # sym_fn = Path("cur_out").resolve()
                sym_fn = self.out_dir / "cur_out"
                try:
                    os.remove(sym_fn)
                except FileNotFoundError:
                    pass

                try:
                    os.symlink(tmp_out_fn, sym_fn)
                    self.log(f"Created symlink in '{sym_fn}'")
                # This may happen if we use a dask scheduler
                except FileExistsError:
                    self.log("Symlink already exists. Skipping generation.")

            # Do at least one cycle. When retries are disabled retry_calc == 0
            # and range(0+1) will result in one cycle
            added_retry_args = False
            for retry in range(self.retry_calc + 1):
                result = subprocess.Popen(
                    args,
                    cwd=path,
                    stdout=handle,
                    stderr=subprocess.PIPE,
                    env=env,
                    shell=shell,
                )
                result.communicate()
                if hasattr(self, 'check_termination'):
                    normal_termination = False
                    try:
                        # Calling check_termination may result in an exception and
                        # normal_termination will stay at False
                        normal_termination = self.check_termination(tmp_out_fn)
                    # The out file may not be present
                    except FileNotFoundError:
                        self.log(
                            f"Could not find out-file {str(tmp_out_fn)} for termination status check!"
                        )
                else:
                    normal_termination = True

                if normal_termination:
                    break
                else:
                    print("Abnormal termination! Retrying calculation.")
                    shutil.copy(tmp_out_fn, str(tmp_out_fn) + f".fail_{retry:02d}")
                    if hasattr(self, 'clean_tmp'):
                        self.clean_tmp(path)
                    else:
                        self.log("'self.clean_tmp()' not implemented!")
                    # Clear tmp_out_fn
                    handle.seek(0)
                    handle.truncate()
                    self.log("Detected abnormal termination! Retrying calculation.")

                    if not added_retry_args:
                        if hasattr(self, 'get_retry_args'):
                            args += self.get_retry_args()
                            added_retry_args = True
                        else:
                            self.log("'self.get_retry_args()' not implemented!")

        # Parse results for desired quantities
        try:
            if run_after:
                self.run_after(path)
            parser_kwargs = {} if parser_kwargs is None else parser_kwargs
            results = self.parser_funcs[calc](path, **parser_kwargs)
            if keep:
                self.keep(path)

        except Exception as err:
            print("Crashed input:")
            print(inp)
            backup_dir = Path(os.getcwd()) / f"crashed_{self.name}"
            self.backup_dir = backup_dir
            if backup_dir.exists():
                shutil.rmtree(backup_dir)
            shutil.copytree(path, backup_dir)
            print(
                f"Copied contents of\n\t'{path}'\nto\n\t'{backup_dir}'.\n"
                "Consider checking the log files there.\n"
            )
            raise err
        finally:
            if (not hold) and self.clean_after:
                self.clean(path)
                if inc_counter:
                    self.calc_counter += 1

        self.path_already_prepared = None
        self.last_run_path = path
        return results

    def run_after(self, path):
        """Meant to be extended.

        This method is called after a calculation was done, but before
        entering ``self.keep()`` and ``self.clean()``. Can be used to call
        tools like formchk or ricctools.
        """

    def popen(self, cmd, cwd=None):
        if isinstance(cmd, str):
            cmd = cmd.split()
        proc = subprocess.Popen(
            cmd,
            cwd=cwd,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        proc.wait()
        return proc

    def prepare_pattern(self, raw_pat):
        """Prepare globs.

        Transforms an entry of ``self.to_keep`` into a glob and a key
        suitable for the use in ``self.keep()``.

        Parameters
        ----------
        raw_pat : str
            Entry of ``self.to_keep``

        Returns
        -------
        pattern : str
            Glob that can be used in Path.glob()
        multi : bool
            Flag if glob may match multiple files.
        key : str
            A key to be used in the ``kept_fns`` dict.
        """
        key_given = None
        if ":" in raw_pat:
            key_given, raw_pat = raw_pat.split(":")
        # Indicates if multiple files are expected
        multi = "*" in raw_pat
        # Drop '*' as it just indicates if we expect multiple matches
        raw_pat = raw_pat.replace("*", "")
        # Interpret it as prefix and drop the two underscores
        if raw_pat.startswith("__"):
            pattern = f"{raw_pat[2:]}*"
            pattern_key = f"{raw_pat[2:]}s"
        # Use raw_pat as suffix
        else:
            pattern = f"*{raw_pat}"
            pattern_key = f"{raw_pat}"
        if key_given:
            pattern_key = key_given
        pattern_key = pattern_key.lower()
        return pattern, multi, pattern_key

    def build_set_plans(self, _set_plans=None):
        if _set_plans is None:
            _set_plans = self._set_plans

        set_plans = list()
        for sp in _set_plans:
            if type(sp) == SetPlan:
                set_plans.append(sp)
                continue
            if type(sp) == dict:
                set_plans.append(SetPlan(**sp))
                continue
            if type(sp) == str:
                sp = (sp,)
            set_plans.append(SetPlan(*sp))
        self.set_plans = set_plans

    def apply_keep_kind(self):
        assert self.keep_kind in KeepKind

        # Nothing to do when set to ALL
        delete_calc_counter = list()
        if self.keep_kind == KeepKind["LATEST"]:
            delete_calc_counter.extend(
                [
                    self.calc_counter - 1,
                ]
                if self.calc_counter > 0
                else []
            )
        elif self.keep_kind == KeepKind["NONE"]:
            delete_calc_counter.extend(list(self.kept_history.keys()))

        for cycle in delete_calc_counter:
            self.log(f"Deleting kept files from cycle {cycle}.")
            for k, f in self.kept_history[cycle].items():
                if type(f) is list:
                    [os.remove(f_) for f_ in f]
                else:
                    os.remove(f)
            del self.kept_history[cycle]

    def keep(self, path):
        """Backup calculation results.

        Parameters
        ----------
        path : Path
            Temporary directory of the calculation.

        Returns
        -------
        kept_fns : dict
            Dictonary holding the filenames that were backed up. The keys
            correspond to the type of file.
        """

        kept_fns = dict()
        for raw_pattern in self.to_keep:
            pattern, multi, key = self.prepare_pattern(raw_pattern)
            globbed = natsorted(path.glob(pattern))
            if not multi:
                assert len(globbed) <= 1, (
                    f"Expected at most one file "
                    f"matching {pattern} in {path}. Found {len(globbed)} "
                    f"files instead ({', '.join([g.name for g in globbed])})!"
                )
            else:
                # I wonder if this has to be a list? Probably to support 'multi', when the pattern
                # contains a '*' character.
                kept_fns[key] = list()
            for tmp_fn in globbed:
                base = tmp_fn.name
                new_fn = self.make_fn(base)
                shutil.copy(tmp_fn, new_fn)
                if multi:
                    kept_fns[key].append(new_fn)
                else:
                    kept_fns[key] = new_fn
        self.kept_history[self.calc_counter] = kept_fns
        self.apply_keep_kind()
        try:
            kept_fns = self.kept_history[self.calc_counter]
        except KeyError:
            kept_fns = dict()
        self.apply_set_plans(kept_fns)
        return kept_fns

    def apply_set_plans(self, kept_fns, set_plans=None):
        if set_plans is None:
            set_plans = self.set_plans

        for sp in set_plans:
            if not sp.condition(self):
                continue
            try:
                setattr(self, sp.name, kept_fns[sp.key])
            except KeyError:
                if sp.fail is not None:
                    self.log(sp.fail(sp.key))

    def clean(self, path):
        """Delete the temporary directory.

        Parameters
        ----------
        path : Path
            Directory to delete.
        """
        shutil.rmtree(path)
        self.log(f"Cleaned {path}")

    def get_restart_info(self):
        """Return a dict containing chkfiles.

        Returns
        -------
        restart_info : dict
            Dictionary holding the calculator state. Used for restoring calculaters
            in restarted calculations.
        """
        try:
            # Convert possible Paths to str
            chkfiles = {k: str(v) for k, v in self.get_chkfiles().items()}
        except AttributeError:
            chkfiles = dict()

        restart_info = {
            "base_name": self.base_name,
            "calc_number": self.calc_number,
            "calc_counter": self.calc_counter,
            "chkfiles": chkfiles,
        }

        return restart_info

    def verify_chkfiles(self, chkfiles):
        """Checks if given chkfiles exist and return them as Paths

        Parameters
        ----------
        chkfiles : dict
            Dictionary holding the chkfiles. The keys correspond to the attribute
            names, the values are strs holding the (potentially full) filename (path).

        Returns
        -------
        paths : dict
            Dictionary of Paths.
        """
        paths = {}
        for key, chkfile in chkfiles.items():
            chkfile = Path(chkfile)
            # If the chkfile exists at the given path we use it as it is.
            if not chkfile.exists():
                self.log(
                    f"Given chkfile '{chkfile}' could not be found! Dropping "
                    "absolute part and trying only its name."
                )
                # Check if relative path exists. This may happen if the calculation
                # has been moved to a different folder.
                name = Path(chkfile.name)
                if name.exists():
                    chkfile = name
                else:
                    self.log(f"'{name}' could not be found! Skipping this chkfile.")
                    continue
            paths[key] = chkfile
        return paths

    def set_restart_info(self, restart_info):
        """Sets restart information (chkfiles etc.) on the calculator.

        Parameters
        -------
        restart_info : dict
            Dictionary holding the calculator state. Used for restoring calculaters
            in restarted calculations.
        """
        try:
            chkfiles = self.verify_chkfiles(restart_info.pop("chkfiles"))
            self.set_chkfiles(chkfiles)
        except KeyError:
            self.log("No chkfiles preset in restart_info")
        except AttributeError:
            self.log(
                "Found chkfiles on restart_info, but 'set_chkfiles' is not "
                "implemented for Calculator."
            )

        self.log("Setting restart_info")
        for key, value in restart_info.items():
            setattr(self, key, value)
            self.log(f"\t{key}: {value}")

    def print_capabilities(self):
        print(
            f"    Can retry?: {hasattr(self, 'check_termination')}\n"
            f"Can track ES??: {hasattr(self, 'prepare_overlap_data')}\n"
        )



"""CLI utilities for standardized exception handling and shared boilerplate."""

from __future__ import annotations

import gc
import logging
import os
import shutil
import sys
import textwrap
import traceback
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Tuple, Type

import click

from .utils import deep_update, load_yaml_dict

logger = logging.getLogger(__name__)


# ---------------------------------------------------------------------------
# YAML source resolution
# ---------------------------------------------------------------------------

def resolve_yaml_sources(
    config_yaml: Optional[Path],
    override_yaml: Optional[Path],
    args_yaml_legacy: Optional[Path],
) -> Tuple[Optional[Path], Optional[Path], bool]:
    """Resolve which YAML files to use, raising on conflicting options."""
    if override_yaml is not None and args_yaml_legacy is not None:
        raise click.BadParameter(
            "Use a single YAML source option."
        )
    if args_yaml_legacy is not None:
        return config_yaml, args_yaml_legacy, True
    return config_yaml, override_yaml, False


def load_merged_yaml_cfg(
    config_yaml: Optional[Path],
    override_yaml: Optional[Path],
) -> Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any]]:
    """Load and merge YAML config and override files.

    Returns ``(merged, config_dict, override_dict)`` so that callers can
    use the individual layers for staged ``apply_yaml_overrides`` and the
    merged dict for ``show_config`` display without re-reading the files.
    """
    config_dict = load_yaml_dict(config_yaml)
    override_dict = load_yaml_dict(override_yaml)
    merged: Dict[str, Any] = {}
    deep_update(merged, config_dict)
    deep_update(merged, override_dict)
    return merged, config_dict, override_dict


# ---------------------------------------------------------------------------
# File helpers
# ---------------------------------------------------------------------------

def link_or_copy_file(src: Path, dst: Path) -> bool:
    """Create a symlink when possible; fall back to copy."""
    try:
        if dst.exists() or dst.is_symlink():
            if dst.is_dir():
                return False
            dst.unlink()
        rel = os.path.relpath(src, start=dst.parent)
        dst.symlink_to(rel)
        return True
    except Exception:
        try:
            shutil.copy2(src, dst)
            return True
        except Exception:
            return False


def run_cli(
    fn: Callable[[], None],
    *,
    label: str,
    zero_step_exc: Optional[Type[BaseException]] = None,
    zero_step_msg: Optional[str] = None,
    opt_exc: Optional[Type[BaseException]] = None,
    opt_msg: Optional[str] = None,
) -> None:
    """Standard CLI exception handling with consistent messaging."""
    try:
        fn()
    except KeyboardInterrupt:
        click.echo("Interrupted by user.", err=True)
        sys.exit(130)
    except Exception as e:
        if zero_step_exc is not None and isinstance(e, zero_step_exc):
            click.echo(
                zero_step_msg
                or "ERROR: Proposed step length dropped below the minimum allowed (ZeroStepLength).",
                err=True,
            )
            sys.exit(2)
        if opt_exc is not None and isinstance(e, opt_exc):
            msg = opt_msg or "ERROR: Optimization failed - {exc}"
            click.echo(msg.format(exc=e), err=True)
            sys.exit(3)
        tb = "".join(traceback.format_exception(type(e), e, e.__traceback__))
        click.echo(
            f"Unhandled error during {label}:\n" + textwrap.indent(tb, "  "),
            err=True,
        )
        sys.exit(1)
    finally:
        # Release GPU memory (model + Hessian) after CLI command finishes
        # so that subsequent pipeline stages (e.g. tsopt → irc) don't OOM.
        # gc.collect() breaks cyclic refs inside torch.nn.Module.
        gc.collect()
        try:
            import torch
            if torch.cuda.is_available():
                torch.cuda.empty_cache()
        except ImportError:
            pass



import numpy as np


def lincs_closure(geom, constraints, order=4):
    """
    Drop conn_nums and keep conns_ in a list of list, instead of an array.
    We could do everything with
        for i, conn_constr in enumerate(conns)
            pass
    instead of the pseudo-code way as given in the paper.
    """
    # Number of constraints
    K = len(constraints)
    # Inverse atom masses
    inv_masses = 1 / geom.masses

    # List of constraint sets
    constraint_sets = [set(c) for c in constraints]
    constraints = np.array(constraints, dtype=int)
    # Indices of the atoms that make up the bond constraints
    atom_1, atom_2 = constraints.T

    # Original bond lengths
    lengths = np.linalg.norm(geom.coords3d[atom_1] - geom.coords3d[atom_2], axis=1)

    # Determine number of constraints that are connected to every constraint
    conn_nums = np.zeros(K, dtype=int)
    conns_ = [list() for _ in range(K)]
    for i, con_1 in enumerate(constraint_sets):
        for j, con_2 in enumerate(constraint_sets):
            if i == j:
                continue

            if con_1 & con_2:
                conn_nums[i] += 1
                conns_[i].append(j)
    # Maximum number of connected constraints
    c_max = conn_nums.max()
    S_diag = 1 / np.sqrt(inv_masses[atom_1] + inv_masses[atom_2])

    # Array containing the indices of the connected constraints for every constraint
    conns = np.zeros((K, c_max), dtype=int)
    for i, c in enumerate(conns_):
        conns[i, :len(c)] = c

    coefs = np.zeros((K, c_max))
    signs = {
        True: -1,
        False: 1,
    }
    for i, con_1 in enumerate(constraint_sets):
        for j in range(c_max):
            conns_ij = conns[i, j]
            sign = signs[(atom_1[i] == atom_1[conns_ij]) or (atom_2[i] == atom_2[conns_ij])]
            # Connected atom
            c = tuple(con_1 & constraint_sets[conns_ij])
            assert len(c) == 1
            c = int(c[0])
            coefs[i, j] = sign * inv_masses[c] * S_diag[i] * S_diag[conns_ij]

    def solve(rhs, sol, new_coords3d, A, B):
        w = 1
        for _ in range(order):
            for i, _ in enumerate(constraints):
                rhs[w, i] = 0.
                for j in range(conn_nums[i]):
                    # rhs[w, i] += A[i, j] * rhs[2-w, conns_[i][j]]
                    rhs[w, i] += A[i, j] * rhs[2-w, conns[i, j]]
                sol[i] += rhs[w, i]
            w = 2 - w

        for i, (a1, a2) in enumerate(constraints):
            new_coords3d[a1] -= inv_masses[a1] * S_diag[i] * sol[i] * B[i].sum()
            new_coords3d[a2] += inv_masses[a2] * S_diag[i] * sol[i] * B[i].sum()

    def lincs(prev_coords3d, new_coords3d):
        # Calculate constraint directions
        B = prev_coords3d[atom_1] - prev_coords3d[atom_2]
        B /= np.linalg.norm(B, axis=1)[:, None]
        print("B", B)

        A = np.zeros((K, c_max))
        rhs = np.zeros((2, K))
        sol = np.zeros(K)

        for i, (a1, a2) in enumerate(constraints):
            print(i, a1, a2)
            for j, k in enumerate(conns[i]):
                A[i, j] = coefs[i, j] * (B[i] * B[k]).sum()
            
            rhs[0, i] = S_diag[i] * ((B[i] * (new_coords3d[a1] - new_coords3d[a2])).sum() - lengths[i])
            sol[i] = rhs[0, i]
        solve(rhs, sol, new_coords3d, A, B)

        # Correction for rotational lenghthening
        for i, (a1, a2) in enumerate(constraints):
            length_i = lengths[i]
            p = (2 * length_i**2 - (-(new_coords3d[a1] - new_coords3d[a2])**2).sum())**0.5
            rhs[0, i] = S_diag[i] * (length_i - p)
            sol[i] = rhs[0, i]

        solve(rhs, sol, new_coords3d, A, B)
        # import pdb; pdb.set_trace()
        return new_coords3d
    return lincs



# [1] https://doi.org/10.1063/1.5090303
#     Geodesic interpolation for reaction pathways
#     Zhu, Thompson, Martinez, 2019

import numpy as np

from pysisyphus.Geometry import Geometry
from pysisyphus.interpolate.Interpolator import Interpolator
from pysisyphus.constants import BOHR2ANG

try:
    from geodesic_interpolate.interpolation import redistribute
    from geodesic_interpolate.geodesic import Geodesic as Geo

    can_geodesic = True
except ImportError:
    can_geodesic = False


class Geodesic(Interpolator):
    def __init__(
        self,
        *args,
        align: bool = False,
        tol: float = 2e-3,
        scaling: float = 1.7,
        dist_cutoff: float = 3.0,
        friction: float = 1e-2,
        maxiter: int = 15,
        microiter: int = 20,
        **kwargs,
    ):
        """Geodesic Interpolation for Reaction Pathways.

        Requires the 'geodesic-interpolate' package found at

            https://github.com/virtualzx-nad/geodesic-interpolate.git

        Parameters
        ----------
        align : bool, optional
            Whether to align geometries, by default False
        tol : float, optional
            Convergence tolerance, by default 2e-3
        scaling : float, optional
            Exponential parameter for morse potential, by default 1.7
        dist_cutoff : float, optional
            Cut-off value for the distance between a pair of atoms
            to be included in the coordinate system, by default 3.0
        friction : float, optional
            Size of friction term used to prevent very large
            change of geometry, by default 1e-2
        maxiter : int, optional
            Maximum number of minimization iterations, by default 15
        microiter : int, optional
            Maximum number of micro iterations for
            sweeping algorithm, by default 20
        """
        super().__init__(*args, align=align, **kwargs)
        self.tol = tol
        self.scaling = scaling
        self.dist_cutoff = dist_cutoff
        self.friction = friction
        self.maxiter = maxiter
        self.microiter = microiter

    def interpolate(self, initial_geom, final_geom, **kwargs):
        if not can_geodesic:
            raise ModuleNotFoundError(
                "Geodesic interpolation requires the geodesic_interpolate package."
            )
        coords3d = np.array((initial_geom.coords3d, final_geom.coords3d))
        coords3d *= BOHR2ANG

        nimages = 2 + self.between
        raw = redistribute(self.atoms, coords3d, nimages, tol=self.tol * 5)

        smoother = Geo(
            self.atoms,
            raw,
            self.scaling,
            threshold=self.dist_cutoff,
            friction=self.friction,
        )
        if len(self.atoms) > 35:
            path = smoother.sweep(
                tol=self.tol, max_iter=self.maxiter, micro_iter=self.microiter
            )
        else:
            path = smoother.smooth(tol=self.tol, max_iter=self.maxiter)
        path /= BOHR2ANG
        interpolated_geoms = [Geometry(self.atoms, coords) for coords in path]
        return interpolated_geoms[1:-1]



from typing import List

import numpy as np
from numpy.typing import NDArray

from pysisyphus.elem_data import COVALENT_RADII as CR
from pysisyphus.intcoords.exceptions import DifferentPrimitivesException
from pysisyphus.intcoords.RedundantCoords import RedundantCoords
from pysisyphus.intcoords.setup import get_bond_sets, BOND_FACTOR
from pysisyphus.intcoords.Stretch import Stretch


def get_tangent(prims1, prims2, dihedral_inds, normalize=False):
    """Normalized tangent between primitive internal coordinates.

    Tangent pointing from prims1 to prims2 in primitive
    internal coordinates, taking into account the periodicity of
    dihedral angles.

    Parameters
    ----------
    prims1 : np.array
        1d-array of primitive internal coordinates in the order
        (stretches, bends, dihedrals).
    prims2 : np.array
        See prims1.
    dihedral_inds : list of int
        Dihedral indices in prims1 and prims2.

    Returns
    -------
    tangent : np.array
        1d array containing the normalized tangent pointing from prims1 to prims2.
    """
    try:
        tangent = prims2 - prims1
    except ValueError:
        raise DifferentPrimitivesException
    diheds = tangent[dihedral_inds].copy()
    diheds_plus = diheds.copy() + 2 * np.pi
    diheds_minus = diheds.copy() - 2 * np.pi
    bigger = np.abs(diheds) > np.abs(diheds_plus)
    diheds[bigger] = diheds_plus[bigger]
    bigger = np.abs(diheds) > np.abs(diheds_minus)
    diheds[bigger] = diheds_minus[bigger]
    tangent[dihedral_inds] = diheds
    if normalize:
        tangent /= np.linalg.norm(tangent)
    return tangent


def get_step(geom, coords):
    assert len(geom.coords) == len(coords)

    if geom.coord_type == "cart":
        diff = geom.coords - coords
    elif geom.coord_type in ("redund", "dlc"):
        diff = -get_tangent(
            geom.internal.prim_coords, coords, geom.internal.dihedral_inds
        )
    else:
        raise Exception("Invalid coord_type!")

    # Convert to DLC
    if geom.coord_type == "dlc":
        diff = geom.internal.U.T.dot(diff)

    return diff


def merge_coordinate_definitions(geom1, geom2):
    typed_prims1 = geom1.internal.typed_prims
    typed_prims2 = geom2.internal.typed_prims
    union = list(set(typed_prims1) | set(typed_prims2))
    union.sort()
    # Check if internal coordinates that are only present in one of the two
    # geometries are valid in the other. If not, we omit these primitives.
    redundant = RedundantCoords(geom1.atoms, geom1.cart_coords, typed_prims=union)
    valid_typed_prims = redundant.typed_prims
    return valid_typed_prims


def form_coordinate_union(geom1, geom2):
    def not_cartesian(geom):
        return geom.coord_type != "cart"

    assert not_cartesian(geom1) and not_cartesian(geom2)
    # The first call yields all primitives from geom1 that are also valid at geom2.
    typed_prims1 = merge_coordinate_definitions(geom1, geom2)
    # The second call yields all primitives from geom2 that are also valid at geom1.
    typed_prims2 = merge_coordinate_definitions(geom2, geom1)

    intersection = list(set(typed_prims1) & set(typed_prims2))
    intersection.sort()
    return intersection


def get_weighted_bond_mode(weighted_bonds, coords3d, remove_translation=True):
    bond_mode = np.zeros_like(coords3d.flatten())
    for *indices, weight in weighted_bonds:
        _, grad = Stretch._calculate(coords3d, indices, gradient=True)
        """
        The gradient gives us the direction into which the bond increases, but
        we want that positive weights correspond to bond formation (distance
        decrease) and negative weights to bond breaking (distance increase),
        so we reverse the sign of the weight.
        """
        bond_mode += -weight * grad

    if remove_translation:
        bm3d = bond_mode.reshape(-1, 3)
        bond_mode = (bm3d - bm3d.mean(axis=0)[None, :]).flatten()

    bond_mode /= np.linalg.norm(bond_mode)
    return bond_mode


def get_weighted_bond_mode_getter(
    target_weighted_bonds, bond_factor=1.2, fractional=False
):
    """Create input for intcoords.helpers.get_weighted_bond_mode.

    Compared to the rest of pysisyphus this method uses a slightly lowered
    bond factor, so it is more strict regarding what is considered a bond
    and what not."""

    def func(atoms, coords3d):
        bond_sets = [
            set(bond)
            for bond in get_bond_sets(atoms, coords3d, bond_factor=bond_factor).tolist()
        ]
        bonds = list()
        frac_weights = list()
        for tbond in target_weighted_bonds:
            *inds, weight = tbond
            assert weight != 0.0
            tset = set(inds)
            """
            Skip target bond if weight is positive (bond is to be formed) and
            bond is already present. Similarly, skip when the bond is to be broken
            and it is not present.
            """
            if (weight > 0 and tset in bond_sets) or (
                weight < 0 and tset not in bond_sets
            ):
                continue
            bonds.append(tbond)

            if fractional:
                from_, to_ = inds
                ref_len = CR[atoms[from_].lower()] + CR[atoms[to_].lower()]
                act_len = np.linalg.norm(coords3d[from_] - coords3d[to_])
                fw = (act_len / ref_len) ** weight
                frac_weights.append(fw)

        if fractional:
            frac_weights = np.array(frac_weights)
            frac_weights /= frac_weights.sum()
            frac_weights = frac_weights ** 2
            frac_weights /= frac_weights.max()
            bonds = [(*inds, fw) for ((*inds, _), fw) in zip(bonds, frac_weights)]
        return bonds

    return func


def get_bond_difference(geom1, geom2, bond_factor=BOND_FACTOR):
    """Return formed and broken bonds when going from geom1 to geom2."""

    assert geom1.atoms == geom2.atoms

    def as_sets(geom):
        bonds = get_bond_sets(geom.atoms, geom.coords3d, bond_factor=bond_factor)
        return set([frozenset(b) for b in bonds.tolist()])

    def as_lists(bonds):
        return [list(b) for b in bonds]

    bonds1 = as_sets(geom1)
    bonds2 = as_sets(geom2)
    formed = as_lists(bonds2 - bonds1)
    broken = as_lists(bonds1 - bonds2)
    return formed, broken


def verbose_bond_difference(formed, broken, key1, key2, atoms=None):
    def atom_bonds(bonds):
        if len(bonds) == 0:
            ab = ""
        elif atoms is not None:
            ab = [f"{atoms[from_]}{from_}-{atoms[to_]}{to_}" for from_, to_ in bonds]
        else:
            ab = bonds
        return ab

    def pos(bonds):
        """Plural or singular?"""
        return ("", "is") if (len(bonds) == 1) else ("s", "are")

    sf, verbf = pos(formed)
    sb, verbb = pos(broken)

    # return (
    # f"{key1}->{key2}: "
    # f"{len(formed)} bond{sf} formed {atom_bonds(formed)}, "
    # f"{len(broken)} bond{sb} broken {atom_bonds(broken)}."
    # )
    return (
        f"{len(formed)} bond{sf} formed {atom_bonds(formed)}",
        f"{len(broken)} bond{sb} broken {atom_bonds(broken)}",
    )


def get_bond_differences_verbose(
    geom1, geom2, bond_factor=BOND_FACTOR, key1="geom1", key2="geom2"
):
    formed, broken = get_bond_difference(geom1, geom2, BOND_FACTOR)
    fverb, bverb = verbose_bond_difference(
        formed, broken, key1, key2, atoms=geom1.atoms
    )
    return formed, broken, fverb, bverb


def interfragment_distance(
    frag1: List[int], frag2: List[int], coords3d: NDArray
) -> float:
    def mean_coords(frag):
        frag_coords = coords3d[frag]
        mean = frag_coords.mean(axis=0)
        return mean

    frag1_mean = mean_coords(frag1)
    frag2_mean = mean_coords(frag2)
    interfrag_dist = float(np.linalg.norm(frag1_mean - frag2_mean))
    return interfrag_dist



# [1] https://doi.org/10.1016/0009-2614(91)90115-P
#     Helgaker, 1991


import numpy as np

from scipy.optimize import newton
from pysisyphus.tsoptimizers.TSHessianOptimizer import TSHessianOptimizer


class TRIM(TSHessianOptimizer):

    def optimize(self):
        energy, gradient, H, eigvals, eigvecs, resetted = self.housekeeping()
        self.update_ts_mode(eigvals, eigvecs)

        self.log(f"Signs of eigenvalue and -vector of root(s) {self.roots} "
                  "will be reversed!")
        # Transform gradient to basis of eigenvectors
        gradient_ = eigvecs.T.dot(gradient)

        # Construct image function by inverting the signs of the eigenvalue and
        # -vector of the mode to follow uphill.
        eigvals_ = eigvals.copy()
        eigvals_[self.roots] *= -1
        gradient_ = gradient_.copy()
        gradient_[self.roots] *= -1

        def get_step(mu):
            zetas = -gradient_ / (eigvals_ - mu)
            # Replace nan with 0.
            zetas = np.nan_to_num(zetas)
            # Transform to original basis
            step = eigvecs * zetas
            step = step.sum(axis=1)
            return step

        def get_step_norm(mu):
            return np.linalg.norm(get_step(mu))

        def func(mu):
            return get_step_norm(mu) - self.trust_radius

        mu = 0
        norm0 = get_step_norm(mu)
        if norm0 > self.trust_radius:
            mu, res = newton(func, x0=mu, full_output=True)
            assert res.converged
            self.log(f"Using levelshift of μ={mu:.4f}")
        else:
            self.log("Took pure newton step without levelshift")

        step = get_step(mu)
        step_norm = np.linalg.norm(step)
        self.log(f"norm(step)={step_norm:.6f}")

        self.predicted_energy_changes.append(self.quadratic_model(gradient, self.H, step))

        return step



# [1] https://www.sciencedirect.com/science/article/pii/000926149500646L
#     Lindh, 1995
# [2] https://pubs.acs.org/doi/pdf/10.1021/j100203a036
#     Fischer, Almlöf, 1992
# [3] https://onlinelibrary.wiley.com/doi/full/10.1002/qua.21049
#     Swart, Bickelhaupt, 2006
# [4] http://dx.doi.org/10.1063/1.4952956
#     Lee-Ping Wang 2016

from math import exp
import itertools as it
from typing import Literal, Optional

import h5py
import numpy as np
from numpy.typing import ArrayLike
from scipy.spatial.distance import pdist, squareform

from pysisyphus.Geometry import Geometry
from pysisyphus.intcoords.PrimTypes import PrimTypes as PT, Bonds, Bends, Dihedrals
from pysisyphus.intcoords.setup import get_pair_covalent_radii
from pysisyphus.io.hessian import save_hessian

import torch

HessInit = Literal[
    "calc", "unit", "fischer", "lindh", "simple", "swart", "xtb", "xtb1", "xtbff"
]


# See [4], last sentences of III.
CART_F = 0.05
# Default force constants
DEFAULT_F = {
    PT.BOND: 0.5,
    PT.AUX_BOND: 0.1,
    PT.HYDROGEN_BOND: 0.1,
    PT.INTERFRAG_BOND: 0.1,
    PT.AUX_INTERFRAG_BOND: 0.05,
    PT.BEND: 0.2,
    PT.LINEAR_BEND: 0.1,
    PT.LINEAR_BEND_COMPLEMENT: 0.1,
    PT.PROPER_DIHEDRAL: 0.1,
    PT.IMPROPER_DIHEDRAL: 0.1,
    PT.OUT_OF_PLANE: 0.1,
    PT.LINEAR_DISPLACEMENT: 0.1,
    PT.LINEAR_DISPLACEMENT_COMPLEMENT: 0.1,
    PT.TRANSLATION_X: CART_F,
    PT.TRANSLATION_Y: CART_F,
    PT.TRANSLATION_Z: CART_F,
    PT.ROTATION_A: CART_F,
    PT.ROTATION_B: CART_F,
    PT.ROTATION_C: CART_F,
    PT.CARTESIAN_X: CART_F,
    PT.CARTESIAN_Y: CART_F,
    PT.CARTESIAN_Z: CART_F,
    PT.BONDED_FRAGMENT: CART_F,
    PT.DUMMY_TORSION: 0.1,
    PT.DISTANCE_FUNCTION: 0.1,
    PT.DUMMY_IMPROPER: 0.1,
}


def simple_guess(geom):
    """Default force constants."""
    h_diag = [DEFAULT_F[type_] for type_, *_ in geom.internal.typed_prims]
    return np.diag(h_diag)


def improved_guess(geom, bond_func, bend_func, dihedral_func):
    H_guess = simple_guess(geom)
    for i, (pt, *indices) in enumerate(geom.internal.typed_prims):
        if pt in Bonds:
            f_func = bond_func
        elif pt in Bends:
            f_func = bend_func
        elif pt in Dihedrals:
            f_func = dihedral_func
        else:
            continue
        try:
            new_f = f_func(indices)
        except ValueError:
            new_f = DEFAULT_F[pt]
        H_guess[i, i] = new_f
    return H_guess


def fischer_guess(geom):
    cdm = pdist(geom.coords3d)
    pair_cov_radii = get_pair_covalent_radii(geom.atoms)

    # For the dihedral force constants we also have to count the number
    # of bonds formed with the centrals atoms of the dihedral.
    central_atoms = [inds[1:3] for inds in geom.internal.dihedral_atom_indices]
    bond_factor = geom.internal.bond_factor
    bond_mat = squareform(cdm <= (pair_cov_radii * bond_factor))
    tors_atom_bonds = dict()
    for a, b in central_atoms:
        # Substract 2, as we don't want the bond between a and b,
        # but this bond will be in both rows of the bond_mat.
        bond_sum = bond_mat[a].sum() + bond_mat[b].sum() - 2
        tors_atom_bonds[(a, b)] = bond_sum

    dist_mat = squareform(cdm)
    pair_cov_radii_mat = squareform(pair_cov_radii)

    def h_bond(indices):
        a, b = indices[:2]
        r_ab = dist_mat[a, b]
        r_ab_cov = pair_cov_radii_mat[a, b]
        return 0.3601 * exp(-1.944 * (r_ab - r_ab_cov))

    def h_bend(indices):
        b, a, c = indices
        r_ab = dist_mat[a, b]
        r_ac = dist_mat[a, c]
        r_ab_cov = pair_cov_radii_mat[a, b]
        r_ac_cov = pair_cov_radii_mat[a, c]
        return 0.089 + 0.11 / (r_ab_cov * r_ac_cov) ** (-0.42) * exp(
            -0.44 * (r_ab + r_ac - r_ab_cov - r_ac_cov)
        )

    def h_dihedral(indices):
        c, a, b, d = indices
        r_ab = dist_mat[a, b]
        r_ab_cov = pair_cov_radii_mat[a, b]
        bond_sum = max(tors_atom_bonds[(a, b)], 0)
        return 0.0015 + 14.0 * bond_sum ** 0.57 / (r_ab * r_ab_cov) ** 4.0 * exp(
            -2.85 * (r_ab - r_ab_cov)
        )

    H = improved_guess(
        geom, bond_func=h_bond, bend_func=h_bend, dihedral_func=h_dihedral
    )
    return H


def lindh_style_guess(geom, ks, rhos):
    """Approximate force constants according to Lindh.[1]

    Bonds:       k_ij = k_r * rho_ij
    Bends:      k_ijk = k_b * rho_ij * rho_jk
    Dihedrals: k_ijkl = k_d * rho_ij * rho_jk * rho_kl
    """

    def k_func(indices):
        rho_product = 1
        inds_len = len(indices)
        for i, ind in enumerate(indices[:-1], 1):
            i1, i2 = ind, indices[i]
            rho_product *= rhos[i1, i2]
        k = ks[inds_len] * rho_product
        return k

    H = improved_guess(geom, bond_func=k_func, bend_func=k_func, dihedral_func=k_func)
    return H


def get_lindh_alpha(atom1, atom2):
    first_period = "h", "he"
    if (atom1 in first_period) and (atom2 in first_period):
        return 1.0
    elif (atom1 in first_period) or (atom2 in first_period):
        return 0.3949
    else:
        return 0.28


def lindh_guess(geom):
    """Slightly modified Lindh model hessian as described in [1].

    Instead of using the tabulated r_ref,ij values from [1] we will use the
    'true' covalent radii as pyberny. The tabulated r_ref,ij value for two
    carbons (2nd period) is 2.87 Bohr. Carbons covalent radius is ~ 1.44 Bohr,
    so 2*1.44 Bohr = 2.88 Bohr which fits nicely with the tabulate value.
    If values for elements > 3rd are requested the alpha values for the 3rd
    period will be (re)used.
    """

    atoms = [a.lower() for a in geom.atoms]
    alphas = [get_lindh_alpha(a1, a2) for a1, a2 in it.combinations(atoms, 2)]
    pair_cov_radii = get_pair_covalent_radii(geom.atoms)
    cdm = pdist(geom.coords3d)
    rhos = squareform(np.exp(alphas * (pair_cov_radii ** 2 - cdm ** 2)))

    ks = {
        2: 0.45,  # Stretches/bonds
        3: 0.15,  # Bends/angles
        4: 0.005,  # Torsions/dihedrals
    }
    return lindh_style_guess(geom, ks, rhos)


def swart_guess(geom):
    pair_cov_radii = get_pair_covalent_radii(geom.atoms)
    cdm = pdist(geom.coords3d)
    rhos = squareform(np.exp(-cdm / pair_cov_radii + 1))
    ks = {
        2: 0.35,
        3: 0.15,
        4: 0.005,
    }
    return lindh_style_guess(geom, ks, rhos)


def xtb_hessian(geom, gfn=None):
    from pysisyphus.calculators.XTB import XTB  # Lazy import
    calc = geom.calculator
    xtb_kwargs = {"charge": calc.charge, "mult": calc.mult, "pal": calc.pal}
    if gfn is not None:
        xtb_kwargs["gfn"] = gfn
    xtb_calc = XTB(**xtb_kwargs)
    geom_ = geom.copy()
    geom_.set_calculator(xtb_calc)
    return geom_.hessian


def get_guess_hessian(
    geometry: Geometry,
    hessian_init: HessInit,
    int_gradient: Optional[ArrayLike] = None,
    cart_gradient: Optional[ArrayLike] = None,
    h5_fn: Optional[str] = None,
):
    """Obtain/calculate (model) Hessian.

    For hessian_init="calc" the Hessian will be in the coord_type
    of the geometry, otherwise a Hessian in primitive internals will
    be returned.
    """
    model_hessian = hessian_init in ("fischer", "lindh", "simple", "swart")
    target_coord_type = geometry.coord_type

    # Recreate geometry with internal coordinates if needed
    if model_hessian and (geometry.coord_type == "cart"):
        geometry = geometry.copy(coord_type="redund")

    hess_funcs = {
        # Calculate true hessian
        "calc": lambda: (geometry.hessian, "calculated exact"),
        # Unit hessian
        "unit": lambda: (np.eye(geometry.coords.size), "unit"),
        # Fischer model hessian
        "fischer": lambda: (fischer_guess(geometry), "Fischer"),
        # Lindh model hessian
        "lindh": lambda: (lindh_guess(geometry), "Lindh"),
        # Simple (0.5, 0.2, 0.1) model hessian
        "simple": lambda: (simple_guess(geometry), "simple"),
        # Swart model hessian
        "swart": lambda: (swart_guess(geometry), "Swart"),
        # XTB hessian using GFN-2
        "xtb": lambda: (xtb_hessian(geometry, gfn=2), "GFN2-XTB"),
        # XTB hessian using GFN-1
        "xtb1": lambda: (xtb_hessian(geometry, gfn=1), "GFN1-XTB"),
        # XTB hessian using GFN-FF
        "xtbff": lambda: (xtb_hessian(geometry, gfn="ff"), "GFN-FF"),
    }
    try:
        H, hess_str = hess_funcs[hessian_init]()
    except KeyError:
        # Only cartesian hessians can be loaded
        if str(hessian_init).endswith(".h5"):
            with h5py.File(hessian_init, "r") as handle:
                cart_hessian = handle["hessian"][:]
        # CFOUR Hessians in the form we need are always named "FCMFINAL"
        elif str(hessian_init).endswith("FCMFINAL"):
            raw_cart_hessian = np.loadtxt(hessian_init, skiprows=1)
            nindices = int(np.sqrt(raw_cart_hessian.size))
            cart_hessian = raw_cart_hessian.reshape((nindices, nindices), order='C')
        else:
            cart_hessian = np.loadtxt(hessian_init)
        geometry.cart_hessian = cart_hessian
        # Use the previously set hessian in whatever coordinate system we
        # actually employ.
        H = geometry.hessian
        hess_str = "saved"

    # if (h5_fn is not None) and (hessian_init == "calc"):
    #     save_hessian(h5_fn, geometry)

    if model_hessian and target_coord_type == "cart":
        if cart_gradient is not None:
            int_gradient = geometry.internal.transform_forces(cart_gradient)
        H = geometry.internal.backtransform_hessian(H, int_gradient=int_gradient)

    return H, hess_str


def ts_hessian(hessian, coord_inds, damp=0.25):
    """According to [3]"""

    inds = list(coord_inds)

    # Use a copy as diag returns only a read-only view
    diag = np.diag(hessian).copy()

    # Reverse sign of reaction coordinates and damp them
    diag[inds] = -1 * damp * diag[inds]
    ts_hess = np.diag(diag)

    # Set off-diagonal elements
    for i, j in it.combinations(inds, 2):
        # fi = force_constants[i]
        # fj = force_constants[j]
        fi = diag[i]
        fj = diag[j]
        f = -((2 * fi * fj) ** 0.5)
        ts_hess[i, j] = f
        ts_hess[j, i] = f
    return ts_hess
